# coding: utf8

"""
This module contains SPM utilities.
"""

INDEX_TISSUE_MAP = {
    1: 'graymatter',
    2: 'whitematter',
    3: 'csf',
    4: 'bone',
    5: 'softtissue',
    6: 'background'
}


def check_spm_home():
    """Check and get SPM_HOME environment variable if present."""
    import os
    import platform
    from colorama import Fore
    from .check_dependency import check_environment_variable
    from .exceptions import ClinicaMissingDependencyError

    spm_home = check_environment_variable('SPM_HOME', 'SPM')

    spm_standalone_home = os.environ.get('SPMSTANDALONE_HOME', '')
    if spm_standalone_home:
        if not os.path.isdir(spm_standalone_home):
            raise ClinicaMissingDependencyError(
                '%s\n[Error] The SPMSTANDALONE_HOME environment variable you gave is not a folder (content: %s).%s'
                % (Fore.RED, spm_standalone_home, Fore.RESET))
        if platform.system() == 'Darwin':
            spm_home = os.path.join(spm_standalone_home, 'spm12.app', 'Contents', 'MacOS', 'spm12_mcr')
        else:
            spm_home = os.path.join(spm_standalone_home, 'spm12_mcr')

    return spm_home


def get_tpm():
    """
    Get Tissue Probability Map (TPM) from SPM
    Args
        None
    Returns
        TPM.nii from SPM
    """
    import os
    from glob import glob
    from os.path import join

    spm_home = os.getenv('SPM_HOME')

    if not spm_home:
        # Try MCR to get a hint on SPM location
        spm_home = os.getenv('SPMSTANDALONE_HOME')

    if not spm_home:
        raise RuntimeError('Could not determine location of your SPM installation. '
                           + 'Neither $SPM_HOME or $SPMSTANDALONE_HOME are present '
                             'in your environment')

    tpm_file_glob = glob(join(spm_home, '**/TPM.nii'), recursive=True)
    if len(tpm_file_glob) == 0:
        raise RuntimeError('No file found for TPM.nii in your $SPM_HOME in ' + spm_home)
    if len(tpm_file_glob) > 1:
        error_str = 'Multiple files found for TPM.nii in your SPM_HOME ' + spm_home + ':'
        for file in tpm_file_glob:
            error_str += '\n\t' + file
        raise RuntimeError(error_str)
    return tpm_file_glob[0]


def spm_standalone_is_available():
    """
        Tells if SPM standalone can be used
    Returns:
        True if SPM standalone is detected, False otherwise. Note that it does not guarentee that SPM (classical) is
        up and running in the system.
    """
    import os
    from os.path import isdir, expandvars

    use_spm_stand = False
    if all(elem in os.environ.keys() for elem in ['SPMSTANDALONE_HOME', 'MCR_HOME']):
        if isdir(expandvars('$SPMSTANDALONE_HOME')) and isdir(expandvars('$MCR_HOME')):
            use_spm_stand = True
        else:
            raise FileNotFoundError('[Error] $SPMSTANDALONE_HOME and $MCR_HOME are defined, but linked to non existent folder')
    return use_spm_stand


def use_spm_standalone():
    """
    Use SPM Standalone with MATLAB Common Runtime
    """
    import os
    from colorama import Fore
    import platform
    from clinica.utils.stream import cprint
    from nipype.interfaces import spm
    # This section of code determines whether to use SPM standalone or not
    if all(elem in os.environ.keys() for elem in ['SPMSTANDALONE_HOME', 'MCR_HOME']):
        spm_standalone_home = os.getenv('SPMSTANDALONE_HOME')
        mcr_home = os.getenv('MCR_HOME')
        if os.path.exists(spm_standalone_home) and os.path.exists(mcr_home):
            cprint(Fore.GREEN + 'SPM standalone has been found and will be used in this pipeline' + Fore.RESET)
            if platform.system().lower().startswith('darwin'):
                matlab_cmd = ('cd %s && ./run_spm12.sh %s script' %
                              (spm_standalone_home, mcr_home))
            elif platform.system().lower().startswith('linux'):
                matlab_cmd = ('%s %s script' %
                              (os.path.join(spm_standalone_home, 'run_spm12.sh'), mcr_home))
            else:
                raise SystemError('Clinica only support macOS and Linux')
            spm.SPMCommand.set_mlab_paths(matlab_cmd=matlab_cmd, use_mcr=True)
            cprint("Using SPM standalone version %s" % spm.SPMCommand().version)
        else:
            raise FileNotFoundError('$SPMSTANDALONE_HOME and $MCR_HOME are defined, but linked to non existent folder ')



# coding: utf8

"""T1 Linear - Clinica Visualizer.
This file has been generated automatically by the `clinica generate template`
command line tool. See here for more details:
http://clinica.run/doc/InteractingWithClinica/
"""



# coding: utf-8

"""
Command line for adni_to_bids converter
"""

import clinica.engine as ce


class AdniToBidsCLI(ce.CmdParser):
    def define_name(self):
        """Define the sub-command name to run this command."""
        self._name = 'adni-to-bids'

    def define_description(self):
        """Define a description of this command."""
        self._description = 'Convert ADNI (http://adni.loni.usc.edu/) into BIDS'

    def define_options(self):
        """Define the sub-command arguments."""
        self._args.add_argument("dataset_directory",
                                help='Path to the ADNI images directory.')
        self._args.add_argument("clinical_data_directory",
                                help='Path to the ADNI clinical data directory.')
        self._args.add_argument("bids_directory",
                                help='Path to the BIDS directory.')
        self._args.add_argument("-c", "--clinical_data_only", action='store_true',
                                help='(Optional) Given the path to an already existing ADNI BIDS folder, convert only '
                                     'the clinical data.')
        self._args.add_argument("-sl", "--subjects_list",
                                help='(Optional) A path to a .txt file containing a list of subject to convert '
                                     '(one for each row).')
        self._args.add_argument("-m", "--modalities", nargs='+',
                                default=['T1', 'PET_FDG', 'PET_AMYLOID', 'PET_TAU', 'DWI', 'FLAIR', 'fMRI'],
                                choices=['T1', 'PET_FDG', 'PET_AMYLOID', 'PET_TAU', 'DWI', 'FLAIR', 'fMRI'],
                                help='(Optional) Convert only the list of selected modalities. '
                                     'By default all modalities are converted. Modalities available: '
                                      'T1, PET_FDG, PET_AMYLOID, PET_TAU, DWI, FLAIR, fMRI.')

    def run_command(self, args):
        from clinica.iotools.converters.adni_to_bids.adni_to_bids import AdniToBids

        adni_to_bids = AdniToBids()

        # Check dcm2nii and dcm2niix dependencies
        adni_to_bids.check_adni_dependencies()

        if not args.clinical_data_only:
            adni_to_bids.convert_images(self.absolute_path(args.dataset_directory),
                                        self.absolute_path(args.clinical_data_directory),
                                        self.absolute_path(args.bids_directory),
                                        args.subjects_list,
                                        args.modalities)

        adni_to_bids.convert_clinical_data(self.absolute_path(args.clinical_data_directory),
                                           self.absolute_path(args.bids_directory))



# coding: utf8

import clinica.engine as ce


class T1VolumeExistingTemplateCLI(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipeline."""
        self._name = 't1-volume-existing-template'

    def define_description(self):
        """Define a description of this pipeline."""
        self._description = ('Volume-based processing of T1-weighted MR images using an existing DARTEL template:\n'
                             'http://clinica.run/doc/Pipelines/T1_Volume/')

    def define_options(self):
        """Define the sub-command arguments."""
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES
        # Clinica compulsory arguments (e.g. BIDS, CAPS, group_id)
        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("bids_directory",
                                  help='Path to the BIDS directory.')
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')
        clinica_comp.add_argument("group_id",
                                  help='User-defined identifier for the provided group of subjects.')
        # Optional arguments (e.g. FWHM)
        optional = self._args.add_argument_group(PIPELINE_CATEGORIES['OPTIONAL'])
        optional.add_argument("-s", "--smooth",
                              nargs='+', type=int, default=[8],
                              help="A list of integers specifying the different isomorphic FWHM in millimeters "
                                   "to smooth the image (default: --smooth 8).")
        # Clinica standard arguments (e.g. --n_procs)
        self.add_clinica_standard_arguments()
        # Advanced arguments (i.e. tricky parameters)
        advanced = self._args.add_argument_group(PIPELINE_CATEGORIES['ADVANCED'])
        # t1-volume-tissue-segmentation
        advanced.add_argument("-tc", "--tissue_classes",
                              metavar='', nargs='+', type=int, default=[1, 2, 3], choices=range(1, 7),
                              help="Tissue classes (1: gray matter (GM), 2: white matter (WM), "
                                   "3: cerebrospinal fluid (CSF), 4: bone, 5: soft-tissue, 6: background) to save "
                                   "(default: GM, WM and CSF i.e. --tissue_classes 1 2 3).")
        advanced.add_argument("-tpm", "--tissue_probability_maps",
                              metavar='TissueProbabilityMap.nii', default=None,
                              help='Tissue probability maps to use for segmentation '
                                   '(default: TPM.nii from SPM software).')
        advanced.add_argument("-dswu", "--dont_save_warped_unmodulated",
                              action='store_true', default=False,
                              help="Do not save warped unmodulated images for tissues specified "
                                   "in --tissue_classes flag.")
        advanced.add_argument("-swm", "--save_warped_modulated",
                              action='store_true', default=False,
                              help="Save warped modulated images for tissues specified in --tissue_classes flag.")
        # t1-volume-tissue-segmentation / t1-volume-create-dartel
        advanced.add_argument("-dt", "--dartel_tissues",
                              metavar='', nargs='+', type=int, default=[1, 2, 3], choices=range(1, 7),
                              help='Tissues to use for DARTEL template calculation '
                                   '(default: GM, WM and CSF i.e. --dartel_tissues 1 2 3).')
        # t1-volume-dartel2mni
        advanced.add_argument("-t", "--tissues",
                              metavar='', nargs='+', type=int, default=[1, 2, 3], choices=range(1, 7),
                              help='Tissues to create flow fields to DARTEL template '
                                   '(default: GM, WM and CSF i.e. --tissues 1 2 3).')
        advanced.add_argument("-m", "--modulate",
                              type=bool, default=True,
                              metavar=('True/False'),
                              help='A boolean. Modulate output images - no modulation preserves concentrations '
                                   '(default: --modulate True).')
        advanced.add_argument("-vs", "--voxel_size",
                              metavar=('float'),
                              nargs=3, type=float,
                              help="A list of 3 floats specifying the voxel sizeof the output image "
                                   "(default: --voxel_size 1.5 1.5 1.5).")

    def run_command(self, args):
        """Run the pipeline with defined args."""
        from colorama import Fore
        from ..t1_volume_tissue_segmentation.t1_volume_tissue_segmentation_cli import T1VolumeTissueSegmentationCLI
        from ..t1_volume_register_dartel.t1_volume_register_dartel_cli import T1VolumeRegisterDartelCLI
        from ..t1_volume_dartel2mni.t1_volume_dartel2mni_cli import T1VolumeDartel2MNICLI
        from ..t1_volume_parcellation.t1_volume_parcellation_cli import T1VolumeParcellationCLI
        from clinica.utils.stream import cprint

        cprint(
            'The t1-volume-existing-template pipeline is divided into 4 parts:\n'
            '\t%st1-volume-tissue-segmentation pipeline%s: Tissue segmentation, bias correction and spatial normalization to MNI space\n'
            '\t%st1-volume-register-dartel pipeline%s: Inter-subject registration using an existing DARTEL template\n'
            '\t%st1-volume-dartel2mni pipeline%s: DARTEL template to MNI\n'
            '\t%st1-volume-parcellation pipeline%s: Atlas statistics'
            % (Fore.BLUE, Fore.RESET, Fore.BLUE, Fore.RESET, Fore.BLUE, Fore.RESET, Fore.BLUE, Fore.RESET)
        )

        cprint('%s\nPart 1/4: Running t1-volume-segmentation pipeline%s' % (Fore.BLUE, Fore.RESET))
        tissue_segmentation_cli = T1VolumeTissueSegmentationCLI()
        tissue_segmentation_cli.run_command(args)

        cprint('%s\nPart 2/4: Running t1-volume-register-dartel pipeline%s' % (Fore.BLUE, Fore.RESET))
        register_dartel_cli = T1VolumeRegisterDartelCLI()
        register_dartel_cli.run_command(args)

        cprint('%s\nPart 3/4: Running t1-volume-dartel2mni pipeline%s' % (Fore.BLUE, Fore.RESET))
        dartel2mni_cli = T1VolumeDartel2MNICLI()
        dartel2mni_cli.run_command(args)

        cprint('%s\nPart 4/4: Running t1-volume-parcellation pipeline%s' % (Fore.BLUE, Fore.RESET))
        parcellation_cli = T1VolumeParcellationCLI()
        parcellation_cli.run_command(args)



# coding: utf8


from os import path
import json
from multiprocessing.pool import ThreadPool
import datetime

import numpy as np

import pandas as pd
from sklearn.svm import SVC
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier
from xgboost import XGBClassifier
from sklearn.model_selection import StratifiedKFold
from sklearn.metrics import roc_auc_score
import itertools
from sklearn.multiclass import OneVsOneClassifier, OneVsRestClassifier

from clinica.pipelines.machine_learning import base
import clinica.pipelines.machine_learning.ml_utils as utils

__author__ = "Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Jorge Samper-Gonzalez", "Pascal Lu", "Simona Bottani"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


class DualSVMAlgorithm(base.MLAlgorithm):

    def _launch_svc(self, kernel_train, x_test, y_train, y_test, c):

        if self._algorithm_params['balanced']:
            svc = SVC(C=c, kernel='precomputed', probability=True, tol=1e-6, class_weight='balanced')
        else:
            svc = SVC(C=c, kernel='precomputed', probability=True, tol=1e-6)

        svc.fit(kernel_train, y_train)
        y_hat_train = svc.predict(kernel_train)
        y_hat = svc.predict(x_test)
        proba_test = svc.predict_proba(x_test)[:, 1]
        auc = roc_auc_score(y_test, proba_test)

        return svc, y_hat, auc, y_hat_train

    def _grid_search(self, kernel_train, x_test, y_train, y_test, c):

        _, y_hat, _, _ = self._launch_svc(kernel_train, x_test, y_train, y_test, c)
        res = utils.evaluate_prediction(y_test, y_hat)

        return res['balanced_accuracy']

    def _select_best_parameter(self, async_result):

        c_values = []
        accuracies = []
        for fold in async_result.keys():
            best_c = -1
            best_acc = -1

            for c, async_acc in async_result[fold].items():

                acc = async_acc.get()
                if acc > best_acc:
                    best_c = c
                    best_acc = acc
            c_values.append(best_c)
            accuracies.append(best_acc)

        best_acc = np.mean(accuracies)
        best_c = np.power(10, np.mean(np.log10(c_values)))

        return {'c': best_c, 'balanced_accuracy': best_acc}

    def evaluate(self, train_index, test_index):

        inner_pool = ThreadPool(self._algorithm_params['n_threads'])
        async_result = {}
        for i in range(self._algorithm_params['grid_search_folds']):
            async_result[i] = {}

        outer_kernel = self._kernel[train_index, :][:, train_index]
        y_train = self._y[train_index]

        skf = StratifiedKFold(n_splits=self._algorithm_params['grid_search_folds'], shuffle=True)
        inner_cv = list(skf.split(np.zeros(len(y_train)), y_train))

        for i in range(len(inner_cv)):
            inner_train_index, inner_test_index = inner_cv[i]

            inner_kernel = outer_kernel[inner_train_index, :][:, inner_train_index]
            x_test_inner = outer_kernel[inner_test_index, :][:, inner_train_index]
            y_train_inner, y_test_inner = y_train[inner_train_index], y_train[inner_test_index]

            for c in self._algorithm_params['c_range']:
                async_result[i][c] = inner_pool.apply_async(self._grid_search,
                                                            (inner_kernel, x_test_inner,
                                                             y_train_inner, y_test_inner, c))
        inner_pool.close()
        inner_pool.join()

        best_parameter = self._select_best_parameter(async_result)
        x_test = self._kernel[test_index, :][:, train_index]
        y_train, y_test = self._y[train_index], self._y[test_index]

        _, y_hat, auc, y_hat_train = self._launch_svc(outer_kernel, x_test, y_train, y_test, best_parameter['c'])

        result = dict()
        result['best_parameter'] = best_parameter
        result['evaluation'] = utils.evaluate_prediction(y_test, y_hat)
        result['evaluation_train'] = utils.evaluate_prediction(y_train, y_hat_train)
        result['y_hat'] = y_hat
        result['y_hat_train'] = y_hat_train
        result['y'] = y_test
        result['y_train'] = y_train
        result['y_index'] = test_index
        result['x_index'] = train_index
        result['auc'] = auc

        return result

    def apply_best_parameters(self, results_list):

        best_c_list = []
        bal_acc_list = []

        for result in results_list:
            best_c_list.append(result['best_parameter']['c'])
            bal_acc_list.append(result['best_parameter']['balanced_accuracy'])

        # 10^(mean of log10 of best Cs of each fold) is selected
        best_c = np.power(10, np.mean(np.log10(best_c_list)))
        # Mean balanced accuracy
        mean_bal_acc = np.mean(bal_acc_list)

        if self._algorithm_params['balanced']:
            svc = SVC(C=best_c, kernel='precomputed', probability=True, tol=1e-6, class_weight='balanced')
        else:
            svc = SVC(C=best_c, kernel='precomputed', probability=True, tol=1e-6)

        svc.fit(self._kernel, self._y)

        return svc, {'c': best_c, 'balanced_accuracy': mean_bal_acc}

    def save_classifier(self, classifier, output_dir):

        np.savetxt(path.join(output_dir, 'dual_coefficients.txt'), classifier.dual_coef_)
        np.savetxt(path.join(output_dir, 'support_vectors_indices.txt'), classifier.support_)
        np.savetxt(path.join(output_dir, 'intersect.txt'), classifier.intercept_)

    def save_weights(self, classifier, x, output_dir):

        dual_coefficients = classifier.dual_coef_
        sv_indices = classifier.support_

        weighted_sv = dual_coefficients.transpose() * x[sv_indices]
        weights = np.sum(weighted_sv, 0)

        np.savetxt(path.join(output_dir, 'weights.txt'), weights)

        return weights

    def save_parameters(self, parameters_dict, output_dir):
        with open(path.join(output_dir, 'best_parameters.json'), 'w') as f:
            json.dump(parameters_dict, f)

    @staticmethod
    def uses_kernel():
        return True

    @staticmethod
    def get_default_parameters():

        parameters_dict = {'balanced': True,
                           'grid_search_folds': 10,
                           'c_range': np.logspace(-6, 2, 17),
                           'n_threads': 15}

        return parameters_dict


class LogisticReg(base.MLAlgorithm):

    def _launch_logistic_reg(self, x_train, x_test, y_train, y_test, c):

        if self._algorithm_params['balanced']:
            classifier = LogisticRegression(penalty=self._algorithm_params['penalty'], tol=1e-6, C=c,
                                            class_weight='balanced')
        else:
            classifier = LogisticRegression(penalty=self._algorithm_params['penalty'], tol=1e-6, C=c)

        classifier.fit(x_train, y_train)
        y_hat_train = classifier.predict(x_train)
        y_hat = classifier.predict(x_test)
        proba_test = classifier.predict_proba(x_test)[:, 1]
        auc = roc_auc_score(y_test, proba_test)

        return classifier, y_hat, auc, y_hat_train

    def _grid_search(self, x_train, x_test, y_train, y_test, c):

        _, y_hat, _, _ = self._launch_logistic_reg(x_train, x_test, y_train, y_test, c)
        res = utils.evaluate_prediction(y_test, y_hat)

        return res['balanced_accuracy']

    def _select_best_parameter(self, async_result):

        c_values = []
        accuracies = []
        for fold in async_result.keys():
            best_c = -1
            best_acc = -1

            for c, async_acc in async_result[fold].items():

                acc = async_acc.get()
                if acc > best_acc:
                    best_c = c
                    best_acc = acc
            c_values.append(best_c)
            accuracies.append(best_acc)

        best_acc = np.mean(accuracies)
        best_c = np.power(10, np.mean(np.log10(c_values)))

        return {'c': best_c, 'balanced_accuracy': best_acc}

    def evaluate(self, train_index, test_index):

        inner_pool = ThreadPool(self._algorithm_params['n_threads'])
        async_result = {}
        for i in range(self._algorithm_params['grid_search_folds']):
            async_result[i] = {}

        x_train = self._x[train_index]
        y_train = self._y[train_index]

        skf = StratifiedKFold(n_splits=self._algorithm_params['grid_search_folds'], shuffle=True)
        inner_cv = list(skf.split(np.zeros(len(y_train)), y_train))

        for i in range(len(inner_cv)):
            inner_train_index, inner_test_index = inner_cv[i]

            x_train_inner = x_train[inner_train_index]
            x_test_inner = x_train[inner_test_index]
            y_train_inner = y_train[inner_train_index]
            y_test_inner = y_train[inner_test_index]

            for c in self._algorithm_params['c_range']:
                async_result[i][c] = inner_pool.apply_async(self._grid_search,
                                                            (x_train_inner, x_test_inner,
                                                             y_train_inner, y_test_inner, c))
        inner_pool.close()
        inner_pool.join()

        best_parameter = self._select_best_parameter(async_result)
        x_test = self._x[test_index]
        y_test = self._y[test_index]

        _, y_hat, auc, y_hat_train = self._launch_logistic_reg(x_train, x_test, y_train, y_test, best_parameter['c'])

        result = dict()
        result['best_parameter'] = best_parameter
        result['evaluation'] = utils.evaluate_prediction(y_test, y_hat)
        result['evaluation_train'] = utils.evaluate_prediction(y_train, y_hat_train)
        result['y_hat'] = y_hat
        result['y_hat_train'] = y_hat_train
        result['y'] = y_test
        result['y_train'] = y_train
        result['y_index'] = test_index
        result['x_index'] = train_index
        result['auc'] = auc

        return result

    def apply_best_parameters(self, results_list):

        best_c_list = []
        bal_acc_list = []

        for result in results_list:
            best_c_list.append(result['best_parameter']['c'])
            bal_acc_list.append(result['best_parameter']['balanced_accuracy'])

        # 10^(mean of log10 of best Cs of each fold) is selected
        best_c = np.power(10, np.mean(np.log10(best_c_list)))
        # Mean balanced accuracy
        mean_bal_acc = np.mean(bal_acc_list)

        if self._algorithm_params['balanced']:
            classifier = LogisticRegression(C=best_c, penalty=self._algorithm_params['penalty'], tol=1e-6,
                                            class_weight='balanced')
        else:
            classifier = LogisticRegression(C=best_c, penalty=self._algorithm_params['penalty'], tol=1e-6)

        classifier.fit(self._x, self._y)

        return classifier, {'c': best_c, 'balanced_accuracy': mean_bal_acc}

    def save_classifier(self, classifier, output_dir):

        np.savetxt(path.join(output_dir, 'weights.txt'), classifier.coef_.transpose())
        np.savetxt(path.join(output_dir, 'intercept.txt'), classifier.intercept_)

    def save_weights(self, classifier, x, output_dir):

        np.savetxt(path.join(output_dir, 'weights.txt'), classifier.coef_.transpose())
        return classifier.coef_.transpose()

    def save_parameters(self, parameters_dict, output_dir):

        with open(path.join(output_dir, 'best_parameters.json'), 'w') as f:
            json.dump(parameters_dict, f)

    @staticmethod
    def _centered_normalised_data(features):
        std = np.std(features, axis=0)
        std[np.where(std == 0)[0]] = 1.
        mean = np.mean(features, axis=0)
        features_bis = (features - mean)/std
        return features_bis, mean, std

    @staticmethod
    def uses_kernel():
        return False

    @staticmethod
    def get_default_parameters():
        parameters_dict = {'penalty': 'l2',
                           'balanced': False,
                           'grid_search_folds': 10,
                           'c_range': np.logspace(-6, 2, 17),
                           'n_threads': 15}

        return parameters_dict


class RandomForest(base.MLAlgorithm):

    def _launch_random_forest(self, x_train, x_test, y_train, y_test, n_estimators, max_depth, min_samples_split,
                              max_features):

        if self._algorithm_params['balanced']:
            classifier = RandomForestClassifier(n_estimators=n_estimators, max_depth=max_depth,
                                                min_samples_split=min_samples_split, max_features=max_features,
                                                class_weight='balanced', n_jobs=self._algorithm_params['n_threads'])
        else:
            classifier = RandomForestClassifier(n_estimators=n_estimators, max_depth=max_depth,
                                                min_samples_split=min_samples_split, max_features=max_features,
                                                n_jobs=self._algorithm_params['n_threads'])

        classifier.fit(x_train, y_train)
        y_hat_train = classifier.predict(x_train)
        y_hat = classifier.predict(x_test)
        proba_test = classifier.predict_proba(x_test)[:, 1]
        auc = roc_auc_score(y_test, proba_test)

        return classifier, y_hat, auc, y_hat_train

    def _grid_search(self, x_train, x_test, y_train, y_test, n_estimators, max_depth, min_samples_split, max_features):

        _, y_hat, _, _ = self._launch_random_forest(x_train, x_test, y_train, y_test,
                                                    n_estimators, max_depth,
                                                    min_samples_split, max_features)
        res = utils.evaluate_prediction(y_test, y_hat)

        return res['balanced_accuracy']

    def _select_best_parameter(self, async_result):

        params_list = []
        accuracies = []

        all_params_acc = []

        for fold in async_result.keys():
            best_params = None
            best_acc = -1

            for params, async_acc in async_result[fold].items():
                acc = async_acc.get()
                if acc > best_acc:
                    best_params = params
                    best_acc = acc

                all_params_acc.append(pd.DataFrame({'n_estimators': params[0],
                                                    'max_depth': params[1],
                                                    'min_samples_split': params[2],
                                                    'max_features': params[3],
                                                    'balanced_accuracy': acc}, index=['i', ]))

            params_list.append(best_params)
            accuracies.append(best_acc)

        best_acc = np.mean(accuracies)
        best_n_estimators = int(round(np.mean([x[0] for x in params_list])))
        best_max_depth = int(round(np.mean([x[1] if x[1] is not None else 50 for x in params_list])))
        best_min_samples_split = int(round(np.mean([x[2] for x in params_list])))

        def max_feature_to_float(m):
            if type(m) is float:
                return m
            if type(m) is int:
                return float(m) / float(self._x.shape[1])
            if m == 'auto' or m == 'sqrt':
                return np.sqrt(self._x.shape[1]) / float(self._x.shape[1])
            if m == 'log2':
                return np.log2(self._x.shape[1]) / float(self._x.shape[1])
            raise ValueError('Not valid value for max_feature: %s' % m)

        float_max_feat = [max_feature_to_float(x[3]) for x in params_list]
        best_max_features = np.mean(float_max_feat)

        return {'n_estimators': best_n_estimators,
                'max_depth': best_max_depth,
                'min_samples_split': best_min_samples_split,
                'max_features': best_max_features,
                'balanced_accuracy': best_acc}

    def evaluate(self, train_index, test_index):

        inner_pool = ThreadPool(self._algorithm_params['n_threads'])
        async_result = {}
        for i in range(self._algorithm_params['grid_search_folds']):
            async_result[i] = {}

        x_train = self._x[train_index]
        y_train = self._y[train_index]

        skf = StratifiedKFold(n_splits=self._algorithm_params['grid_search_folds'], shuffle=True)
        inner_cv = list(skf.split(np.zeros(len(y_train)), y_train))

        parameters_combinations = list(itertools.product(self._algorithm_params['n_estimators_range'],
                                                         self._algorithm_params['max_depth_range'],
                                                         self._algorithm_params['min_samples_split_range'],
                                                         self._algorithm_params['max_features_range']))

        for i in range(len(inner_cv)):
            inner_train_index, inner_test_index = inner_cv[i]

            x_train_inner = x_train[inner_train_index]
            x_test_inner = x_train[inner_test_index]
            y_train_inner = y_train[inner_train_index]
            y_test_inner = y_train[inner_test_index]

            for parameters in parameters_combinations:
                async_result[i][parameters] = inner_pool.apply_async(self._grid_search,
                                                                     (x_train_inner, x_test_inner,
                                                                      y_train_inner, y_test_inner,
                                                                      parameters[0], parameters[1],
                                                                      parameters[2], parameters[3]))
        inner_pool.close()
        inner_pool.join()
        best_parameter = self._select_best_parameter(async_result)
        x_test = self._x[test_index]
        y_test = self._y[test_index]

        _, y_hat, auc, y_hat_train = self._launch_random_forest(x_train, x_test, y_train, y_test,
                                                                best_parameter['n_estimators'],
                                                                best_parameter['max_depth'],
                                                                best_parameter['min_samples_split'],
                                                                best_parameter['max_features'])

        result = dict()
        result['best_parameter'] = best_parameter
        result['evaluation'] = utils.evaluate_prediction(y_test, y_hat)
        result['evaluation_train'] = utils.evaluate_prediction(y_train, y_hat_train)
        result['y_hat'] = y_hat
        result['y_hat_train'] = y_hat_train
        result['y'] = y_test
        result['y_train'] = y_train
        result['y_index'] = test_index
        result['x_index'] = train_index
        result['auc'] = auc

        return result

    def evaluate_no_cv(self, train_index, test_index):

        x_train = self._x[train_index]
        y_train = self._y[train_index]
        x_test = self._x[test_index]
        y_test = self._y[test_index]

        best_parameter = dict()
        best_parameter['n_estimators'] = self._algorithm_params['n_estimators_range']
        best_parameter['max_depth'] = self._algorithm_params['max_depth_range']
        best_parameter['min_samples_split'] = self._algorithm_params['min_samples_split_range']
        best_parameter['max_features'] = self._algorithm_params['max_features_range']

        _, y_hat, auc, y_hat_train = self._launch_random_forest(x_train, x_test, y_train, y_test,
                                                                self._algorithm_params['n_estimators_range'],
                                                                self._algorithm_params['max_depth_range'],
                                                                self._algorithm_params['min_samples_split_range'],
                                                                self._algorithm_params['max_features_range'])
        result = dict()
        result['best_parameter'] = best_parameter
        result['evaluation'] = utils.evaluate_prediction(y_test, y_hat)
        best_parameter['balanced_accuracy'] = result['evaluation']['balanced_accuracy']
        result['evaluation_train'] = utils.evaluate_prediction(y_train, y_hat_train)
        result['y_hat'] = y_hat
        result['y_hat_train'] = y_hat_train
        result['y'] = y_test
        result['y_train'] = y_train
        result['y_index'] = test_index
        result['x_index'] = train_index
        result['auc'] = auc

        return result

    def apply_best_parameters(self, results_list):

        mean_bal_acc = np.mean([result['best_parameter']['balanced_accuracy'] for result in results_list])
        best_n_estimators = int(round(np.mean([result['best_parameter']['n_estimators'] for result in results_list])))
        best_max_depth = int(round(np.mean([result['best_parameter']['max_depth']
                                            if result['best_parameter']['max_depth'] is not None
                                            else 50 for result in results_list])))
        best_min_samples_split = int(round(np.mean([result['best_parameter']['min_samples_split']
                                                    for result in results_list])))

        max_feat = []
        n_features = self._x.shape[1]
        for result in results_list:
            result_feat = result['best_parameter']['max_features']

            if result_feat is None:
                max_features = 1.0
            elif result_feat in ["auto", "sqrt"]:
                max_features = np.sqrt(n_features) / n_features
            elif result_feat == "log2":
                max_features = np.log2(n_features) / n_features
            elif isinstance(result_feat, int):
                max_features = float(result_feat) / n_features
            elif isinstance(result_feat, float):
                max_features = result_feat
            else:
                raise ("Unknown max_features type")

            max_feat.append(max_features)
        best_max_features = np.mean(max_feat)

        if self._algorithm_params['balanced']:
            classifier = RandomForestClassifier(n_estimators=best_n_estimators, max_depth=best_max_depth,
                                                min_samples_split=best_min_samples_split,
                                                max_features=best_max_features,
                                                class_weight='balanced', n_jobs=self._algorithm_params['n_threads'])
        else:
            classifier = RandomForestClassifier(n_estimators=best_n_estimators, max_depth=best_max_depth,
                                                min_samples_split=best_min_samples_split,
                                                max_features=best_max_features,
                                                n_jobs=self._algorithm_params['n_threads'])

        classifier.fit(self._x, self._y)

        return classifier, {'n_estimators': best_n_estimators,
                            'max_depth': best_max_depth,
                            'min_samples_split': best_min_samples_split,
                            'max_features': best_max_features,
                            'balanced_accuracy': mean_bal_acc}

    def save_classifier(self, classifier, output_dir):
        np.savetxt(path.join(output_dir, 'feature_importances.txt'), classifier.feature_importances_)
        # print classifier.estimators_
        # np.savetxt(path.join(output_dir, 'estimators.txt'), str(classifier.estimators_))

    def save_weights(self, classifier, x, output_dir):

        np.savetxt(path.join(output_dir, 'weights.txt'), classifier.feature_importances_)
        return classifier.feature_importances_

    def save_parameters(self, parameters_dict, output_dir):

        with open(path.join(output_dir, 'best_parameters.json'), 'w') as f:
            json.dump(parameters_dict, f)

    @staticmethod
    def uses_kernel():
        return False

    @staticmethod
    def get_default_parameters():

        parameters_dict = {'balanced': False,
                           'grid_search_folds': 10,
                           'n_estimators_range': (10, 25, 50, 100, 150, 200, 500),
                           'max_depth_range': (None, 6, 8, 10, 12),
                           'min_samples_split_range': (2, 4, 6, 8),
                           'max_features_range': ('auto', 0.1, 0.2, 0.3, 0.4, 0.5),
                           'n_threads': 15}

        return parameters_dict


class XGBoost(base.MLAlgorithm):

    def _launch_xgboost(self, x_train, x_test, y_train, y_test, max_depth, learning_rate, n_estimators,
                        colsample_bytree):

        if self._algorithm_params['balanced']:
            # set scale_pos_weight
            # http://xgboost.readthedocs.io/en/latest//how_to/param_tuning.html
            scale_pos_weight = float(len(self._y - sum(self._y)) / sum(self._y))
            classifier = XGBClassifier(max_depth=max_depth, learning_rate=learning_rate, n_estimators=n_estimators,
                                       n_jobs=self._algorithm_params['n_threads'], colsample_bytree=colsample_bytree,
                                       reg_alpha=self._algorithm_params['reg_alpha'],
                                       reg_lambda=self._algorithm_params['reg_lambda'],
                                       scale_pos_weight=scale_pos_weight)
        else:
            classifier = XGBClassifier(max_depth=max_depth, learning_rate=learning_rate, n_estimators=n_estimators,
                                       n_jobs=self._algorithm_params['n_threads'], colsample_bytree=colsample_bytree,
                                       reg_alpha=self._algorithm_params['reg_alpha'],
                                       reg_lambda=self._algorithm_params['reg_lambda'])

        classifier.fit(x_train, y_train)
        y_hat_train = classifier.predict(x_train)
        y_hat = classifier.predict(x_test)
        proba_test = classifier.predict_proba(x_test)[:, 1]
        auc = roc_auc_score(y_test, proba_test)

        return classifier, y_hat, auc, y_hat_train

    def _grid_search(self, x_train, x_test, y_train, y_test, max_depth, learning_rate, n_estimators, colsample_bytree):

        _, y_hat, _, _ = self._launch_xgboost(x_train, x_test, y_train, y_test, max_depth, learning_rate, n_estimators,
                                              colsample_bytree)
        res = utils.evaluate_prediction(y_test, y_hat)

        return res['balanced_accuracy']

    def _select_best_parameter(self, async_result):

        params_list = []
        accuracies = []

        all_params_acc = []

        for fold in async_result.keys():
            best_params = None
            best_acc = -1

            for params, async_acc in async_result[fold].items():
                acc = async_acc.get()
                if acc > best_acc:
                    best_params = params
                    best_acc = acc

                all_params_acc.append(pd.DataFrame({'max_depth': params[0],
                                                    'learning_rate': params[1],
                                                    'n_estimators': params[2],
                                                    'colsample_bytree': params[3],
                                                    'balanced_accuracy': acc}, index=['i', ]))

            params_list.append(best_params)
            accuracies.append(best_acc)

        best_acc = np.mean(accuracies)

        best_max_depth = int(round(np.mean([x[0] for x in params_list])))
        best_learning_rate = np.mean([x[1] for x in params_list])
        best_n_estimators = int(round(np.mean([x[2] for x in params_list])))
        best_colsample_bytree = np.mean([x[3] for x in params_list])

        return {'max_depth': best_max_depth,
                'learning_rate': best_learning_rate,
                'n_estimators': best_n_estimators,
                'colsample_bytree': best_colsample_bytree,
                'balanced_accuracy': best_acc}

    def evaluate(self, train_index, test_index):

        inner_pool = ThreadPool(self._algorithm_params['n_threads'])
        async_result = {}
        for i in range(self._algorithm_params['grid_search_folds']):
            async_result[i] = {}

        x_train = self._x[train_index]
        y_train = self._y[train_index]

        skf = StratifiedKFold(n_splits=self._algorithm_params['grid_search_folds'], shuffle=True)
        inner_cv = list(skf.split(np.zeros(len(y_train)), y_train))

        parameters_combinations = list(itertools.product(self._algorithm_params['max_depth_range'],
                                                         self._algorithm_params['learning_rate_range'],
                                                         self._algorithm_params['n_estimators_range'],
                                                         self._algorithm_params['colsample_bytree_range']))

        for i in range(len(inner_cv)):
            inner_train_index, inner_test_index = inner_cv[i]

            x_train_inner = x_train[inner_train_index]
            x_test_inner = x_train[inner_test_index]
            y_train_inner = y_train[inner_train_index]
            y_test_inner = y_train[inner_test_index]

            for parameters in parameters_combinations:
                async_result[i][parameters] = inner_pool.apply_async(self._grid_search,
                                                                     (x_train_inner, x_test_inner,
                                                                      y_train_inner, y_test_inner,
                                                                      parameters[0], parameters[1],
                                                                      parameters[2], parameters[3]))
        inner_pool.close()
        inner_pool.join()
        best_parameter = self._select_best_parameter(async_result)
        x_test = self._x[test_index]
        y_test = self._y[test_index]

        _, y_hat, auc, y_hat_train = self._launch_xgboost(x_train, x_test, y_train, y_test,
                                                          best_parameter['max_depth'],
                                                          best_parameter['learning_rate'],
                                                          best_parameter['n_estimators'],
                                                          best_parameter['colsample_bytree'])

        result = dict()
        result['best_parameter'] = best_parameter
        result['evaluation'] = utils.evaluate_prediction(y_test, y_hat)
        result['evaluation_train'] = utils.evaluate_prediction(y_train, y_hat_train)
        result['y_hat'] = y_hat
        result['y_hat_train'] = y_hat_train
        result['y'] = y_test
        result['y_train'] = y_train
        result['y_index'] = test_index
        result['x_index'] = train_index
        result['auc'] = auc

        return result

    def evaluate_no_cv(self, train_index, test_index):

        x_train = self._x[train_index]
        y_train = self._y[train_index]
        x_test = self._x[test_index]
        y_test = self._y[test_index]

        best_parameter = dict()
        best_parameter['max_depth'] = self._algorithm_params['max_depth_range']
        best_parameter['learning_rate'] = self._algorithm_params['learning_rate_range']
        best_parameter['n_estimators'] = self._algorithm_params['n_estimators_range']
        best_parameter['colsample_bytree'] = self._algorithm_params['colsample_bytree_range']

        _, y_hat, auc, y_hat_train = self._launch_xgboost(x_train, x_test, y_train, y_test,
                                                          self._algorithm_params['max_depth_range'],
                                                          self._algorithm_params['learning_rate_range'],
                                                          self._algorithm_params['n_estimators_range'],
                                                          self._algorithm_params['colsample_bytree_range'])
        result = dict()
        result['best_parameter'] = best_parameter
        result['evaluation'] = utils.evaluate_prediction(y_test, y_hat)
        best_parameter['balanced_accuracy'] = result['evaluation']['balanced_accuracy']
        result['evaluation_train'] = utils.evaluate_prediction(y_train, y_hat_train)
        result['y_hat'] = y_hat
        result['y_hat_train'] = y_hat_train
        result['y'] = y_test
        result['y_train'] = y_train
        result['y_index'] = test_index
        result['x_index'] = train_index
        result['auc'] = auc

        return result

    def apply_best_parameters(self, results_list):

        mean_bal_acc = np.mean([result['best_parameter']['balanced_accuracy'] for result in results_list])

        best_max_depth = int(round(np.mean([result['best_parameter']['max_depth'] for result in results_list])))
        best_learning_rate = np.mean([result['best_parameter']['learning_rate'] for result in results_list])
        best_n_estimators = int(round(np.mean([result['best_parameter']['n_estimators'] for result in results_list])))
        best_colsample_bytree = np.mean([result['best_parameter']['colsample_bytree'] for result in results_list])

        if self._algorithm_params['balanced']:
            scale_pos_weight = float(len(self._y - sum(self._y)) / sum(self._y))

            classifier = XGBClassifier(max_depth=best_max_depth, learning_rate=best_learning_rate,
                                       n_estimators=best_n_estimators, n_jobs=self._algorithm_params['n_threads'],
                                       colsample_bytree=best_colsample_bytree,
                                       reg_alpha=self._algorithm_params['reg_alpha'],
                                       reg_lambda=self._algorithm_params['reg_lambda'],
                                       scale_pos_weight=scale_pos_weight)
        else:
            classifier = XGBClassifier(max_depth=best_max_depth, learning_rate=best_learning_rate,
                                       n_estimators=best_n_estimators, n_jobs=self._algorithm_params['n_threads'],
                                       colsample_bytree=best_colsample_bytree,
                                       reg_alpha=self._algorithm_params['reg_alpha'],
                                       reg_lambda=self._algorithm_params['reg_lambda'])

        classifier.fit(self._x, self._y)

        return classifier, {'max_depth': best_max_depth,
                            'learning_rate': best_learning_rate,
                            'n_estimators': best_n_estimators,
                            'colsample_bytree': best_colsample_bytree,
                            'balanced_accuracy': mean_bal_acc}

    def save_classifier(self, classifier, output_dir):
        np.savetxt(path.join(output_dir, 'feature_importances.txt'), classifier.feature_importances_)
        # print classifier.estimators_
        # np.savetxt(path.join(output_dir, 'estimators.txt'), str(classifier.estimators_))

    def save_weights(self, classifier, x, output_dir):

        np.savetxt(path.join(output_dir, 'weights.txt'), classifier.feature_importances_)
        return classifier.feature_importances_

    def save_parameters(self, parameters_dict, output_dir):

        with open(path.join(output_dir, 'best_parameters.json'), 'w') as f:
            json.dump(parameters_dict, f)

    @staticmethod
    def uses_kernel():
        return False

    @staticmethod
    def get_default_parameters():
        parameters_dict = {'balanced': False,
                           'grid_search_folds': 10,
                           'max_depth_range': (0, 6),
                           'learning_rate_range': (0.1, 0.3),
                           'n_estimators_range': (100, 200),
                           'colsample_bytree_range': (0.5, 1),
                           'reg_alpha': 0,
                           'reg_lambda': 1,
                           'n_threads': 15}

        return parameters_dict


class OneVsOneSVM(base.MLAlgorithm):

    def _launch_svc(self, kernel_train, x_test, y_train, y_test, c):

        if self._algorithm_params['balanced']:
            svc = OneVsOneClassifier(SVC(C=c, kernel='precomputed', probability=True, tol=1e-6,
                                         class_weight='balanced'))
        else:
            svc = OneVsOneClassifier(SVC(C=c, kernel='precomputed', probability=True, tol=1e-6))

        svc.fit(kernel_train, y_train)
        y_hat_train = svc.predict(kernel_train)
        y_hat = svc.predict(x_test)
        proba_test = svc.predict_proba(x_test)[:, 1]

        return svc, y_hat, y_hat_train

    def _grid_search(self, kernel_train, x_test, y_train, y_test, c):
        # y_hat is the value predicted

        _, y_hat, _ = self._launch_svc(kernel_train, x_test, y_train, y_test, c)
        res = utils.evaluate_prediction_multiclass(y_test, y_hat)

        return res['balanced_accuracy']

    def _select_best_parameter(self, async_result):

        c_values = []
        accuracies = []
        for fold in async_result.keys():
            best_c = -1
            best_acc = -1

            for c, async_acc in async_result[fold].items():

                acc = async_acc.get()
                if acc > best_acc:
                    best_c = c
                    best_acc = acc
            c_values.append(best_c)
            accuracies.append(best_acc)

        best_acc = np.mean(accuracies)
        best_c = np.power(10, np.mean(np.log10(c_values)))

        return {'c': best_c, 'balanced_accuracy': best_acc}

    def evaluate(self, train_index, test_index):

        inner_pool = ThreadPool(self._algorithm_params['n_threads'])
        async_result = {}
        for i in range(self._algorithm_params['grid_search_folds']):
            async_result[i] = {}

        outer_kernel = self._kernel[train_index, :][:, train_index]
        y_train = self._y[train_index]

        skf = StratifiedKFold(n_splits=self._algorithm_params['grid_search_folds'], shuffle=True)
        inner_cv = list(skf.split(np.zeros(len(y_train)), y_train))

        for i in range(len(inner_cv)):
            inner_train_index, inner_test_index = inner_cv[i]

            inner_kernel = outer_kernel[inner_train_index, :][:, inner_train_index]
            x_test_inner = outer_kernel[inner_test_index, :][:, inner_train_index]
            y_train_inner, y_test_inner = y_train[inner_train_index], y_train[inner_test_index]

            for c in self._algorithm_params['c_range']:
                async_result[i][c] = inner_pool.apply_async(self._grid_search,
                                                            (inner_kernel, x_test_inner,
                                                             y_train_inner, y_test_inner, c))
        inner_pool.close()
        inner_pool.join()

        best_parameter = self._select_best_parameter(async_result)
        x_test = self._kernel[test_index, :][:, train_index]
        y_train, y_test = self._y[train_index], self._y[test_index]

        _, y_hat, y_hat_train = self._launch_svc(outer_kernel, x_test, y_train, y_test, best_parameter['c'])

        result = dict()
        result['best_parameter'] = best_parameter
        result['evaluation'] = utils.evaluate_prediction_multiclass(y_test, y_hat)
        result['evaluation_train'] = utils.evaluate_prediction_multiclass(y_train, y_hat_train)
        result['y_hat'] = y_hat
        result['y_hat_train'] = y_hat_train
        result['y'] = y_test
        result['y_train'] = y_train
        result['y_index'] = test_index
        result['x_index'] = train_index

        return result

    def apply_best_parameters(self, results_list):

        best_c_list = []
        bal_acc_list = []

        for result in results_list:
            best_c_list.append(result['best_parameter']['c'])
            bal_acc_list.append(result['best_parameter']['balanced_accuracy'])

        # 10^(mean of log10 of best Cs of each fold) is selected
        best_c = np.power(10, np.mean(np.log10(best_c_list)))
        # Mean balanced accuracy
        mean_bal_acc = np.mean(bal_acc_list)

        if self._algorithm_params['balanced']:
            svc = OneVsOneClassifier(SVC(C=best_c, kernel='precomputed', probability=True, tol=1e-6,
                                         class_weight='balanced'))
        else:
            svc = OneVsOneClassifier(SVC(C=best_c, kernel='precomputed', probability=True, tol=1e-6))

        svc.fit(self._kernel, self._y)

        return svc, {'c': best_c, 'balanced_accuracy': mean_bal_acc}

    def save_classifier(self, classifier, output_dir):

        np.savetxt(path.join(output_dir, 'support_vectors_indices.txt'), classifier.support_)
        np.savetxt(path.join(output_dir, 'intersect.txt'), classifier.intercept_)

    def save_weights(self, classifier, x, output_dir):

        dual_coefficients = classifier.dual_coef_
        sv_indices = classifier.support_

        weighted_sv = dual_coefficients.transpose() * x[sv_indices]
        weights = np.sum(weighted_sv, 0)

        np.savetxt(path.join(output_dir, 'weights.txt'), weights)

        return weights

    def save_parameters(self, parameters_dict, output_dir):
        with open(path.join(output_dir, 'best_parameters.json'), 'w') as f:
            json.dump(parameters_dict, f)

    @staticmethod
    def uses_kernel():
        return True

    @staticmethod
    def get_default_parameters():

        parameters_dict = {'balanced': True,
                           'grid_search_folds': 10,
                           'c_range': np.logspace(-6, 2, 17),
                           'n_threads': 15}

        return parameters_dict


class OneVsRestSVM(base.MLAlgorithm):

    def _launch_svc(self, kernel_train, x_test, y_train, y_test, c):

        if self._algorithm_params['balanced']:
            svc = OneVsRestClassifier(SVC(C=c, kernel='precomputed', probability=True, tol=1e-6,
                                          class_weight='balanced'))
        else:
            svc = OneVsRestClassifier(SVC(C=c, kernel='precomputed', probability=True, tol=1e-6))

        svc.fit(kernel_train, y_train)
        y_hat_train = svc.predict(kernel_train)
        y_hat = svc.predict(x_test)
        proba_test = svc.predict_proba(x_test)[:, 1]

        return svc, y_hat, y_hat_train

    def _grid_search(self, kernel_train, x_test, y_train, y_test, c):
        # y_hat is the value predicted

        _, y_hat, _ = self._launch_svc(kernel_train, x_test, y_train, y_test, c)
        res = utils.evaluate_prediction_multiclass(y_test, y_hat)

        return res['balanced_accuracy']

    def _select_best_parameter(self, async_result):

        c_values = []
        accuracies = []
        for fold in async_result.keys():
            best_c = -1
            best_acc = -1

            for c, async_acc in async_result[fold].items():

                acc = async_acc.get()
                if acc > best_acc:
                    best_c = c
                    best_acc = acc
            c_values.append(best_c)
            accuracies.append(best_acc)

        best_acc = np.mean(accuracies)
        best_c = np.power(10, np.mean(np.log10(c_values)))

        return {'c': best_c, 'balanced_accuracy': best_acc}

    def evaluate(self, train_index, test_index):

        inner_pool = ThreadPool(self._algorithm_params['n_threads'])
        async_result = {}
        for i in range(self._algorithm_params['grid_search_folds']):
            async_result[i] = {}

        outer_kernel = self._kernel[train_index, :][:, train_index]
        y_train = self._y[train_index]

        skf = StratifiedKFold(n_splits=self._algorithm_params['grid_search_folds'], shuffle=True)
        inner_cv = list(skf.split(np.zeros(len(y_train)), y_train))

        for i in range(len(inner_cv)):
            inner_train_index, inner_test_index = inner_cv[i]

            inner_kernel = outer_kernel[inner_train_index, :][:, inner_train_index]
            x_test_inner = outer_kernel[inner_test_index, :][:, inner_train_index]
            y_train_inner, y_test_inner = y_train[inner_train_index], y_train[inner_test_index]

            for c in self._algorithm_params['c_range']:
                async_result[i][c] = inner_pool.apply_async(self._grid_search,
                                                            (inner_kernel, x_test_inner,
                                                             y_train_inner, y_test_inner, c))
        inner_pool.close()
        inner_pool.join()

        best_parameter = self._select_best_parameter(async_result)
        x_test = self._kernel[test_index, :][:, train_index]
        y_train, y_test = self._y[train_index], self._y[test_index]

        _, y_hat, y_hat_train = self._launch_svc(outer_kernel, x_test, y_train, y_test, best_parameter['c'])

        result = dict()
        result['best_parameter'] = best_parameter
        result['evaluation'] = utils.evaluate_prediction_multiclass(y_test, y_hat)
        result['evaluation_train'] = utils.evaluate_prediction_multiclass(y_train, y_hat_train)
        result['y_hat'] = y_hat
        result['y_hat_train'] = y_hat_train
        result['y'] = y_test
        result['y_train'] = y_train
        result['y_index'] = test_index
        result['x_index'] = train_index

        return result

    def apply_best_parameters(self, results_list):

        best_c_list = []
        bal_acc_list = []

        for result in results_list:
            best_c_list.append(result['best_parameter']['c'])
            bal_acc_list.append(result['best_parameter']['balanced_accuracy'])

        # 10^(mean of log10 of best Cs of each fold) is selected
        best_c = np.power(10, np.mean(np.log10(best_c_list)))
        # Mean balanced accuracy
        mean_bal_acc = np.mean(bal_acc_list)

        if self._algorithm_params['balanced']:
            svc = OneVsOneClassifier(SVC(C=best_c, kernel='precomputed', probability=True, tol=1e-6,
                                         class_weight='balanced'))
        else:
            svc = OneVsOneClassifier(SVC(C=best_c, kernel='precomputed', probability=True, tol=1e-6))

        svc.fit(self._kernel, self._y)

        return svc, {'c': best_c, 'balanced_accuracy': mean_bal_acc}

    def save_classifier(self, classifier, output_dir):

        np.savetxt(path.join(output_dir, 'support_vectors_indices.txt'), classifier.support_)
        np.savetxt(path.join(output_dir, 'intersect.txt'), classifier.intercept_)

    def save_weights(self, classifier, x, output_dir):

        dual_coefficients = classifier.dual_coef_
        sv_indices = classifier.support_

        weighted_sv = dual_coefficients.transpose() * x[sv_indices]
        weights = np.sum(weighted_sv, 0)

        np.savetxt(path.join(output_dir, 'weights.txt'), weights)

        return weights

    def save_parameters(self, parameters_dict, output_dir):
        with open(path.join(output_dir, 'best_parameters.json'), 'w') as f:
            json.dump(parameters_dict, f)

    @staticmethod
    def uses_kernel():
        return True

    @staticmethod
    def get_default_parameters():

        parameters_dict = {'balanced': True,
                           'grid_search_folds': 10,
                           'c_range': np.logspace(-6, 2, 17),
                           'n_threads': 15}

        return parameters_dict



# coding: utf8

# Get containers to produce the CAPS structure


def container_from_filename(bids_or_caps_filename):
    """Extract container from BIDS or CAPS file.
    Args:
       bids_or_caps_filename (str): full path to BIDS or CAPS filename.
    Returns:
       Container path of the form "subjects/<participant_id>/<session_id>"
    Examples:
       >>> from clinica.utils.nipype import container_from_filename
       >>> container_from_filename('/path/to/bids/sub-CLNC01/ses-M00/anat/sub-CLNC01_ses-M00_T1w.nii.gz')
               'subjects/sub-CLNC01/ses-M00'
       >>> container_from_filename('caps/subjects/sub-CLNC01/ses-M00/dwi/preprocessing/sub-CLNC01_ses-M00_preproc.nii')
               'subjects/sub-CLNC01/ses-M00'
    """

    import os
    import re
    m = re.search(r'(sub-[a-zA-Z0-9]+)/(ses-[a-zA-Z0-9]+)', bids_or_caps_filename)
    if m is None:
        raise ValueError(
                'Input filename is not in a BIDS or CAPS compliant format.'
                'It does not contain the participant and session ID.'
                )
    subject = m.group(1)
    session = m.group(2)
    return os.path.join('subjects', subject, session)


def fix_join(path, *paths):
    # This workaround is used in pipelines like DWIPreprocessingUsingT1
    # In the workflow.connect part, you can use some function that are used as string, causing an import error
    import os
    return os.path.join(path, *paths)



# coding: utf8


def init_input_node(caps_dir, participant_id, session_id, long_id, output_dir):
    """Initialize the pipeline."""
    import os
    import errno
    import datetime
    import platform
    from tempfile import mkdtemp
    from colorama import Fore
    from clinica.utils.longitudinal import read_sessions
    from clinica.utils.ux import print_begin_image
    from clinica.utils.stream import cprint

    # Extract <image_id>
    image_id = '{0}_{1}_{2}'.format(participant_id, session_id, long_id)

    # Create SUBJECTS_DIR for recon-all (otherwise, the command won't run)
    if platform.system().lower().startswith('darwin'):
        # Special case: On macOS, 'recon-all -long' can failed if the $SUBJECTS_DIR is too long
        # To circumvent this issue, we create a sym link in $(TMP) so that $SUBJECTS_DIR is a short path
        subjects_dir = mkdtemp()
        now = datetime.datetime.now().strftime('%H:%M:%S')
        cprint('%s[%s] Needs to create a $SUBJECTS_DIR folder in %s for %s (macOS case). %s' %
               (Fore.YELLOW, now, subjects_dir, image_id.replace('_', ' | '), Fore.RESET))
    else:
        subjects_dir = os.path.join(output_dir, image_id)

    try:
        os.makedirs(subjects_dir)
    except OSError as e:
        if e.errno != errno.EEXIST:  # EEXIST: folder already exists
            raise e

    # Create symbolic link containing cross-sectional segmentation(s) in SUBJECTS_DIR so that recon-all can run
    for s_id in read_sessions(caps_dir, participant_id, long_id):
        cross_sectional_path = os.path.join(
            caps_dir,
            'subjects',
            participant_id,
            s_id,
            't1',
            'freesurfer_cross_sectional',
            participant_id + '_' + s_id
        )
        os.symlink(cross_sectional_path, os.path.join(subjects_dir, participant_id + '_' + s_id))

    # Create symbolic links containing unbiased template in SUBJECTS_DIR so that recon-all can run
    template_path = os.path.join(
        caps_dir,
        'subjects',
        participant_id,
        long_id,
        'freesurfer_unbiased_template',
        participant_id + '_' + long_id
    )
    os.symlink(template_path, os.path.join(subjects_dir, participant_id + '_' + long_id))

    print_begin_image(image_id)

    return subjects_dir


def run_recon_all_long(subjects_dir,
                       participant_id,
                       session_id,
                       long_id,
                       directive):
    """Run recon-all to create a longitudinal correction of a time point.

    Note:
    Longitudinal correction with FreeSurfer expects arguments to follow this syntax:
    recon-all -long <tpN_id> <template_id> -all; e.g.: recon-all -long sub-CLNC01_ses-M00 sub-CLNC01_long-M00M18 -all

    Currently, Nipype does not provide interface for "recon-all -long" case. As a result, ReconAll interface should be
    modified to handle this situation. In the meantime, the arguments of this function follows ReconAll.inputs names
    namely:
        - "-long <tpN_id> <template_id>" is likely to be fed to ReconAll.inputs.args
        - "-all" will be fed to ReconAll.inputs.directive

    Folder containing the longitudinal correction has the following convention:
    <tpN_id>.long.<template_id>; e.g.: sub-CLNC01_ses-M00.long.sub-CLNC01_long-M00M18
    which is automatically generated by FreeSurfer.

    This folder name is likely to be retrieved in ReconAll.outputs.subject_id.

    See official documentation (https://surfer.nmr.mgh.harvard.edu/fswiki/LongitudinalProcessing) for details.
    """
    import subprocess

    # Prepare arguments for recon-all.
    flags = " -long {0}_{1} {0}_{2} ".format(participant_id, session_id, long_id)

    recon_all_long_command = 'recon-all {0} -sd {1} {2}'.format(flags, subjects_dir, directive)
    subprocess_run_recon_all_long = subprocess.run(
        recon_all_long_command,
        shell=True,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL)
    if subprocess_run_recon_all_long.returncode != 0:
        raise ValueError('recon-all -long failed, returned non-zero code')

    subject_id = '{0}_{1}.long.{0}_{2}'.format(participant_id, session_id, long_id)

    return subject_id


def write_tsv_files(subjects_dir, subject_id):
    """
    Generate statistics TSV files in `subjects_dir`/regional_measures folder for `subject_id`.

    Notes:
        We do not need to check the line "finished without error" in scripts/recon-all.log.
        If an error occurs, it will be detected by Nipype and the next nodes (i.e.
        write_tsv_files will not be called).
    """
    import os
    import datetime
    from colorama import Fore
    from clinica.utils.stream import cprint
    from clinica.utils.freesurfer import generate_regional_measures, extract_image_id_from_longitudinal_segmentation

    image_id = extract_image_id_from_longitudinal_segmentation(subject_id)
    str_image_id = image_id.participant_id + '_' + image_id.session_id + '_' + image_id.long_id
    if os.path.isfile(os.path.join(subjects_dir, subject_id, 'mri', 'aparc+aseg.mgz')):
        generate_regional_measures(subjects_dir, subject_id)
    else:
        now = datetime.datetime.now().strftime('%H:%M:%S')
        cprint('%s[%s] %s does not contain mri/aseg+aparc.mgz file. '
               'Creation of regional_measures/ folder will be skipped.%s' %
               (Fore.YELLOW, now, str_image_id.replace('_', ' | '), Fore.RESET))
    return subject_id


def move_subjects_dir_to_source_dir(subjects_dir, source_dir, subject_id):
    """
    Move content of `subjects_dir`/`subject_id` to `source_dir`.

    This function will move content of `subject_id` if recon-all has run
    in $(TMP). This happens when FreeSurfer is run on macOS. Content of
    $(TMP)/`subject_id` is copied to `source_dir` before the deletion of $(TMP).

    Args:
        subjects_dir: $(TMP), if segmentation was performed on macOS,
            <base_dir>/<Pipeline.Name>/ReconAll/`subject_id` otherwise
        source_dir: <base_dir>/<Pipeline.Name>/ReconAll folder
        subject_id: Subject ID (e.g. "sub-CLNC01_ses-M00.long.sub-CLNC01_long-M00M18")

    Returns:
        subject_id for node connection with Nipype
    """
    import os
    import shutil
    import datetime
    from colorama import Fore
    from clinica.utils.freesurfer import extract_image_id_from_longitudinal_segmentation
    from clinica.utils.stream import cprint

    image_id = extract_image_id_from_longitudinal_segmentation(subject_id)
    participant_id = image_id.participant_id
    session_id = image_id.session_id
    long_id = image_id.long_id
    str_image_id = image_id.participant_id + '_' + image_id.session_id + '_' + image_id.long_id

    if source_dir not in subjects_dir:
        shutil.copytree(
            src=os.path.join(subjects_dir, subject_id),
            dst=os.path.join(source_dir, str_image_id, subject_id),
            symlinks=True
        )
        shutil.copytree(
            src=os.path.join(subjects_dir, 'regional_measures'),
            dst=os.path.join(source_dir, str_image_id, 'regional_measures'),
            symlinks=True
        )
        shutil.rmtree(subjects_dir)
        now = datetime.datetime.now().strftime('%H:%M:%S')
        cprint('%s[%s] Segmentation of %s has moved to working directory and $SUBJECTS_DIR folder (%s) was deleted%s' %
               (Fore.YELLOW, now, subject_id.replace('_', ' | '), subjects_dir, Fore.RESET))

    return subject_id


def save_to_caps(source_dir, subject_id, caps_dir, overwrite_caps=False):
    """Save `source_dir`/`subject_id`/ to CAPS folder.

    This function copies FreeSurfer segmentation and regional_measures folder of `source_dir`/`image_id`/ to
    `caps_dir`/subjects/<participant_id>/<session_id>/<long_id>/t1_freesurfer_longitudinal/
    where `image_id` = <participant_id>_<session_id>_<long_id>.

    The `source_dir`/`image_id`/ folder should contain the following elements:
        - fsaverage, lh.EC_average and rh.EC_average symbolic links automatically generated by recon-all
        - symbolic links to cross-sectional segmentation(s) and unbiased template needed for recon-all
        - <participant_id>_<session_id>.long.<participant_id>_<long_id>/ folder containing the FreeSurfer segmentation
        - regional_measures/ folder containing TSV files

    Notes:
        We do not need to check the line "finished without error" in scripts/recon-all.log.
        If an error occurs, it will be detected by Nipype and the next nodes (i.e. save_to_caps will not be called).
    """
    import os
    import datetime
    import shutil
    from colorama import Fore
    from clinica.utils.stream import cprint
    from clinica.utils.ux import print_end_image
    from clinica.utils.freesurfer import extract_image_id_from_longitudinal_segmentation

    image_id = extract_image_id_from_longitudinal_segmentation(subject_id)
    participant_id = image_id.participant_id
    session_id = image_id.session_id
    long_id = image_id.long_id
    str_image_id = image_id.participant_id + '_' + image_id.session_id + '_' + image_id.long_id

    destination_dir = os.path.join(
        os.path.expanduser(caps_dir),
        'subjects',
        participant_id,
        session_id,
        't1',
        long_id,
        'freesurfer_longitudinal'
    )

    # Save FreeSurfer segmentation
    representative_file = os.path.join(subject_id, 'mri', 'aparc+aseg.mgz')
    representative_source_file = os.path.join(os.path.expanduser(source_dir), str_image_id, representative_file)
    representative_destination_file = os.path.join(destination_dir, representative_file)
    if os.path.isfile(representative_source_file):
        if os.path.isfile(representative_destination_file):
            if overwrite_caps:
                shutil.rmtree(destination_dir)
                shutil.copytree(
                    src=os.path.join(source_dir, subject_id, subject_id),
                    dst=os.path.join(destination_dir, subject_id),
                    symlinks=True
                )
                shutil.copytree(
                    src=os.path.join(source_dir, subject_id, 'regional_measures'),
                    dst=os.path.join(destination_dir, 'regional_measures'),
                    symlinks=True
                )
        else:
            shutil.copytree(
                src=os.path.join(source_dir, str_image_id, subject_id),
                dst=os.path.join(destination_dir, subject_id),
                symlinks=True
            )
            shutil.copytree(
                src=os.path.join(source_dir, str_image_id, 'regional_measures'),
                dst=os.path.join(destination_dir, 'regional_measures'),
                symlinks=True
            )
        print_end_image(str_image_id)
    else:
        now = datetime.datetime.now().strftime('%H:%M:%S')
        cprint('%s[%s] %s does not contain mri/aseg+aparc.mgz file. Copy will be skipped.%s' %
               (Fore.YELLOW, now, str_image_id.replace('_', ' | '), Fore.RESET))
    return str_image_id


def get_processed_images(caps_directory, part_ids, sess_ids, long_ids):
    """
    Extract image IDs (e.g. ["sub-CLNC01_ses-M00_long-M00M18", "sub-CLNC01_ses-M18_long-M00M18"]) of outputs
    already processed by T1FreeSurferLongitudinalCorrection pipeline.
    """
    import os
    image_ids = []
    if os.path.isdir(caps_directory):
        for (participant_id, session_id, long_id) in zip(part_ids, sess_ids, long_ids):
            output_file = os.path.join(
                os.path.expanduser(caps_directory),
                'subjects',
                participant_id,
                session_id,
                't1',
                long_id,
                'freesurfer_longitudinal',
                participant_id + '_' + session_id + '.long.' + participant_id + '_' + long_id,
                'mri',
                'aparc+aseg.mgz'
            )
            if os.path.isfile(output_file):
                image_ids.append(participant_id + '_' + session_id + '_' + long_id)
    return image_ids



# coding: utf8


class Patient(object):
    """Class that handles sessions ordering, image quality hierarchy and final BIDS structure"""

    def __init__(self, name, path_patient, path_ida):
        self.name = name
        self.path = path_patient
        self.path_ida = path_ida

        self.sessions, self.corr_sessions = self.get_sessions()
        self.ses0 = self.get_ses0()

    def get_sessions(self):
        """
        Returns the list of all sessions that the patient attended
        """
        import pandas as pd

        def change_format_date(date):
            sol = ""
            new_date = date.split('/')
            if len(new_date[0]) != 2:
                new_date[0] = '0' + new_date[0]
            if len(new_date[1]) != 2:
                new_date[1] = '0' + new_date[1]
            sol = new_date[2] + '-' + new_date[0] + '-' + new_date[1]
            return sol

        df = pd.read_csv(self.path_ida, sep='\t')
        patient_clinical = df.loc[df['Subject ID'] == self.name]
        dic = dict()
        sessions = list(set(list(patient_clinical['Study Date'])))

        for ses in sessions:
            visit = patient_clinical.loc[patient_clinical['Study Date'] == ses]['Visit'].iloc[0]
            ses_number = visit.split(' ')[1]
            dic[change_format_date(ses)] = int(ses_number)

        return sessions, dic

    def get_ses0(self):
        if self.sessions == []:
            print('Warning: sessions not defined for ' + str(self.name))
            return None
        return self.sessions[0]

    def get_sesNumber(self, ses_date):
        return self.corr_sessions[ses_date]

    def get_sesName(self, ses_date):
        number = str(self.get_sesNumber(ses_date))
        while len(number) < 2:
            number = '0' + number

        res = 'ses-M' + number

        return res

    def get_name(self):
        return 'sub-NIFD' + self.name.replace("_", "")

    def order_sessions(self, equivalences, descriptors, folders):
        """
        Orders sessions of a given patient

        Args:
          folders : List of all paths to medical images

        Returns:
          same_ses : dictionary containing paths ordered by sessions
                    i.e. same_ses = {'session_number' : [all paths to MRIs made during said session]}
        """
        from clinica.iotools.converters.nifd_to_bids.nifd_utils import extract_date

        dates = [extract_date(path) for path in folders]
        dates = list(set(dates))

        same_dates = [(self.get_sesName(date), [fold for fold in folders if date in fold]) for date in dates]

        ses_names = [tupl[0] for tupl in same_dates]
        ses_names = list(set(ses_names))

        same_ses = {}
        for tupl in same_dates:
            if tupl[0] not in same_ses:
                same_ses[tupl[0]] = tupl[1]
            else:
                same_ses[tupl[0]].extend(tupl[1])

        return same_ses

    def order_priorities(self, equivalences, descriptors, folders):
        """
        Order all folders in the BIDS format following the priorities defined in the json file

        Args:
            equivalences:   Data structure of the form :
                            equivalences['medical_image_name'] = (Descriptor_instance, modalityLabel),
                            links a medical name to its descriptor
            descriptors: List of descriptor objects, this list is build from the config_dcm2bids.json file
            folders: List of directories' paths containing Dicom files for a given patient

        Returns:
          sol: dictionary of the form sol['session_id']['dataType']['Priority']['Final_name'] = [paths/to/dcm]
        """
        from clinica.iotools.converters.nifd_to_bids.nifd_utils import extract_name_med_img

        sol = {}
        sessions = self.order_sessions(equivalences, descriptors, folders)

        for ses_id in sessions:
            sol[ses_id] = {}
            for path in sessions[ses_id]:

                med_name = extract_name_med_img(path, equivalences)
                dataType = equivalences[med_name][0].dataType
                priority = equivalences[med_name][0].priority
                final_name = self.get_name() + '_' + ses_id + '_' + equivalences[med_name][0].get_bids_info()

                if dataType not in sol[ses_id]:
                    sol[ses_id][dataType] = {}
                if priority not in sol[ses_id][dataType]:
                    sol[ses_id][dataType][priority] = {}
                if final_name not in sol[ses_id][dataType][priority]:
                    sol[ses_id][dataType][priority][final_name] = [path]
                elif path not in sol[ses_id][dataType][priority][final_name]:
                    sol[ses_id][dataType][priority][final_name].extend([path])

        return sol

    def clean_conflicts(self, equivalences, descriptors, folders, conflicts_manager, pat):
        """
        Order all folders in the BIDS format and removes all conflicts
        The output contains all paths to be used by the converter

        Args:
            equivalences:   Data structure of the form :
                            equivalences['medical_image_name'] = (Descriptor_instance, modalityLabel),
                            links a medical name to its descriptor
            descriptors: List of descriptor objects, this list is build from the config_dcm2bids.json file
            folders: List of directories' paths containing Dicom files for a given patient
            conflicts_manager: Instance of Manage_conflicts(), handles quality of converted images (ex : T1_DIS3D > T1)
            pat: Patient ID

        Returns:
            A data structure of the form : ordered_bids[session][datatype][priority][name]
            ordered_bids[session][datatype][priority][name] ontains a list that is either empty or has a single path to a dicom file.
            All paths will then be converted to Nifti following the BIDS format.
        """
        from clinica.iotools.converters.nifd_to_bids.nifd_utils import extract_name_med_img

        ordered_bids = self.order_priorities(equivalences, descriptors, folders)

        for ses in ordered_bids:
            for dType in ordered_bids[ses]:
                encountered = []
                highest_priority = max(ordered_bids[ses][dType].keys())
                while highest_priority != 0:
                    if highest_priority in ordered_bids[ses][dType]:
                        for name in ordered_bids[ses][dType][highest_priority]:
                            enc = name.split('_')[-1]
                            if enc in encountered:
                                ordered_bids[ses][dType][highest_priority][name] = []
                            else:
                                encountered.append(enc)
                                if len(ordered_bids[ses][dType][highest_priority][name]) > 1:

                                    conflit = [extract_name_med_img(path, equivalences) for path in
                                               ordered_bids[ses][dType][highest_priority][name]]
                                    try:
                                        select = conflicts_manager.make_decision(conflit)
                                    except Exception:
                                        print('Warning : ' + str(
                                            conflit) + ' not in expected conflicts, files will not be converted [Subject ID : ' + pat + ']')
                                        select = None
                                    if select is None:
                                        ordered_bids[ses][dType][highest_priority][name] = []
                                    else:
                                        val = [path for path in ordered_bids[ses][dType][highest_priority][name] if
                                               str(select) in path]
                                        ordered_bids[ses][dType][highest_priority][name] = val
                    highest_priority -= 1

        return ordered_bids

    def get_conflicts(self, equivalences, descriptors, folders, conflicts={}):
        # from nifd_utils import extract_name_med_img
        """
        Returns a list of all possible conflicts in dataset according to the json file
        A conflict happens when several images end up in the same directory with the same name after BIDS conversion
        """
        ordered_bids = self.order_priorities(equivalences, descriptors, folders)

        for ses in ordered_bids:
            for dataType in ordered_bids[ses]:
                highest_priority = max(ordered_bids[ses][dataType].keys())
                name_encountered = []
                while highest_priority != 0:
                    if highest_priority in ordered_bids[ses][dataType]:
                        for name in ordered_bids[ses][dataType][highest_priority]:
                            name2 = name
                            name = name.split('_')[-1]
                            if name not in name_encountered:
                                name_encountered.append(name)
                                if len(ordered_bids[ses][dataType][highest_priority][name2]) > 1:
                                    if name not in conflicts:
                                        conflicts[name] = [ordered_bids[ses][dataType][highest_priority][name2]]
                                    else:
                                        conflicts[name].append(ordered_bids[ses][dataType][highest_priority][name2])

                    highest_priority -= 1

        return conflicts



# coding: utf-8

"""
StatisticsVolumeCorrection - Clinica Utilities.
"""


def peak_correction(t_map, t_threshold, output_name=None):
    """
    Threshold the t_map with t_threshold. Pixel intensities that are less than t_threshold are set to 0, other values
    are left unchanged.

    Args:
        t_map: (str) path to t-statistics nifti map
        t_threshold: (float) threshold on t value
        output_name: (str) optional output name

    Returns:
        path to the generated file.
    """
    import nibabel as nib
    from os.path import join, basename, abspath

    original_nifti = nib.load(t_map)
    data = original_nifti.get_data()
    data[data < t_threshold] = 0
    new_data = nib.Nifti1Image(data, affine=original_nifti.affine, header=original_nifti.header)
    if output_name:
        filename = output_name
    else:
        filename = join('./peak_corrected_' + str(t_threshold) + basename(t_map))
    nib.save(new_data, filename)
    return abspath(filename)


def cluster_correction(t_map, t_thresh, c_thresh, output_name=None):
    """
    Performs cluster correction. First t_map is thresholded with t_thresh (like in peak_correction()). Then, clusters
    that have a size less than c_thresh are removed
    Args:
        t_map: (str) path to t-statistics nifti map
        t_thresh: (float) threshold on t value
        c_thresh: (int) minimal size of clusters after thresholding
        output_name: (str) optional output name

    Returns:
        path to the generated file.
    """
    import nibabel as nib
    from os.path import join, basename, abspath
    import numpy as np
    from scipy.ndimage.measurements import label

    original_nifti = nib.load(t_map)
    data = original_nifti.get_data()
    data[data < t_thresh] = 0
    labeled_mask, num_features = label(data)
    for i in range(1, num_features + 1):
        if np.sum(labeled_mask == i) < c_thresh:
            print('Label number ' + str(i) + ' cluster size is: ' + str(np.sum(labeled_mask == i)) + ' so it is removed')
            data[labeled_mask == i] = 0
    new_data = nib.Nifti1Image(data, affine=original_nifti.affine, header=original_nifti.header)
    if output_name:
        filename = output_name
    else:
        filename = join('./cluster_corrected_t-' + str(t_thresh) + '_c-' + str(c_thresh) + basename(t_map))
    nib.save(new_data, filename)
    return abspath(filename)


def produce_figures(nii_file, template, type_of_correction, t_thresh, c_thresh, n_cuts):
    """
    Produce the output figures

    Args:
        nii_file: (str) path to the nifti file (generated at previous steps)
        template: (str) path to template used for the stat map plot
        type_of_correction: (str) Can be either FWE or FDR (used only in potential figure titles)
        t_thresh: (str) t value threshold used (used only in potential figure titles)
        c_thresh: (int) cluster minimal size used (used only in potential figure titles)
        n_cuts: (int) number of cuts in fig

    Returns:
        List of path to image files: glass brain, statmap along x, statmap along y, statmap along z
    """
    from nilearn import plotting
    import numpy as np
    from os.path import abspath

    assert type_of_correction in ['FWE', 'FDR'], 'Type of correction must be FWE or FDR'
    if not np.isnan(c_thresh):
        correction = 'Cluster'
    else:
        correction = 'Peak'

    my_title = correction + ' correction ' + type_of_correction + ' Threshold = ' + str(t_thresh)
    if not np.isnan(c_thresh):
        my_title = my_title + ' - min cluster size = ' + str(c_thresh),

    plotting.plot_glass_brain(nii_file,
                              output_file='./glass_brain.png')

    plotting.plot_stat_map(nii_file,
                           display_mode='x',
                           cut_coords=np.linspace(-70, 67, n_cuts),
                           bg_img=template,
                           colorbar=False,
                           draw_cross=True,
                           output_file='./statmap_x.png')

    plotting.plot_stat_map(nii_file,
                           display_mode='y',
                           cut_coords=np.linspace(-104, 69, n_cuts),
                           bg_img=template,
                           colorbar=False,
                           draw_cross=True,
                           output_file='./statmap_y.png')

    plotting.plot_stat_map(nii_file,
                           display_mode='z',
                           cut_coords=np.linspace(-45, 78, n_cuts),
                           bg_img=template,
                           colorbar=False,
                           draw_cross=True,
                           output_file='./statmap_z.png')

    return [abspath('./glass_brain.png'),
            abspath('./statmap_x.png'),
            abspath('./statmap_y.png'),
            abspath('./statmap_z.png')]


def generate_output(t_map, figs, name):
    """
        Produce output
    Args:
        t_map: (str) path to t-map on which whole pipeline was based
        figs: (list of str) paths to figs to save
        name: (str) name of the correction (ex: cluster_correction_FWE)

    Returns:
        Nothing
    """
    from os import makedirs
    from os.path import join, dirname, basename, splitext
    from shutil import copyfile

    # Will extract group-GroupTest_AD-lt-CN_measure-fdg_fwhm-8_TStatistics from TStatistics file
    t_map_basename = splitext(basename(t_map))[0]

    out_folder = join(dirname(t_map), t_map_basename.replace('TStatistics', name))
    makedirs(out_folder)
    copyfile(figs[0], join(out_folder, t_map_basename.replace('TStatistics', 'desc-' + name + '_GlassBrain.png')))
    copyfile(figs[1], join(out_folder, t_map_basename.replace('TStatistics', 'desc-' + name + '_axis-x_TStatistics.png')))
    copyfile(figs[2], join(out_folder, t_map_basename.replace('TStatistics', 'desc-' + name + '_axis-y_TStatistics.png')))
    copyfile(figs[3], join(out_folder, t_map_basename.replace('TStatistics', 'desc-' + name + '_axis-z_TStatistics.png')))



# coding: utf8

import os
from os import path
import numpy as np
import pandas as pd
from sklearn.model_selection import StratifiedKFold, StratifiedShuffleSplit
from multiprocessing.pool import ThreadPool

from clinica.pipelines.machine_learning import base

__author__ = "Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Jorge Samper-Gonzalez"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


class KFoldCV(base.MLValidation):

    def validate(self, y):

        if self._validation_params['splits_indices'] is None:
            skf = StratifiedKFold(n_splits=self._validation_params['n_folds'], shuffle=True)
            self._validation_params['splits_indices'] = list(skf.split(np.zeros(len(y)), y))

        async_pool = ThreadPool(self._validation_params['n_threads'])
        async_result = {}

        for i in range(self._validation_params['n_folds']):

            train_index, test_index = self._validation_params['splits_indices'][i]
            async_result[i] = async_pool.apply_async(self._ml_algorithm.evaluate, (train_index, test_index))

        async_pool.close()
        async_pool.join()

        for i in range(self._validation_params['n_folds']):
            self._validation_results.append(async_result[i].get())

        self._classifier, self._best_params = self._ml_algorithm.apply_best_parameters(self._validation_results)

        return self._classifier, self._best_params, self._validation_results

    def save_results(self, output_dir):
        if self._validation_results is None:
            raise Exception("No results to save. Method validate() must be run before save_results().")

        subjects_folds = []
        results_folds = []
        container_dir = path.join(output_dir, 'folds')

        if not path.exists(container_dir):
            os.makedirs(container_dir)

        for i in range(len(self._validation_results)):
            subjects_df = pd.DataFrame({'y': self._validation_results[i]['y'],
                                        'y_hat': self._validation_results[i]['y_hat'],
                                        'y_index': self._validation_results[i]['y_index']})
            subjects_df.to_csv(path.join(container_dir, 'subjects_fold-' + str(i) + '.tsv'),
                               index=False, sep='\t', encoding='utf-8')
            subjects_folds.append(subjects_df)

            results_df = pd.DataFrame({'balanced_accuracy': self._validation_results[i]['evaluation']['balanced_accuracy'],
                                       'auc': self._validation_results[i]['auc'],
                                       'accuracy': self._validation_results[i]['evaluation']['accuracy'],
                                       'sensitivity': self._validation_results[i]['evaluation']['sensitivity'],
                                       'specificity': self._validation_results[i]['evaluation']['specificity'],
                                       'ppv': self._validation_results[i]['evaluation']['ppv'],
                                       'npv': self._validation_results[i]['evaluation']['npv'],
                                       'train_balanced_accuracy': self._validation_results[i]['evaluation_train']['balanced_accuracy'],
                                       'train_accuracy': self._validation_results[i]['evaluation_train']['accuracy'],
                                       'train_sensitivity': self._validation_results[i]['evaluation_train']['sensitivity'],
                                       'train_specificity': self._validation_results[i]['evaluation_train']['specificity'],
                                       'train_ppv': self._validation_results[i]['evaluation_train']['ppv'],
                                       'train_npv': self._validation_results[i]['evaluation_train']['npv']
                                       }, index=['i', ])

            results_df.to_csv(path.join(container_dir, 'results_fold-' + str(i) + '.tsv'),
                              index=False, sep='\t', encoding='utf-8')
            results_folds.append(results_df)

        all_subjects = pd.concat(subjects_folds)
        all_subjects.to_csv(path.join(output_dir, 'subjects.tsv'),
                            index=False, sep='\t', encoding='utf-8')

        all_results = pd.concat(results_folds)
        all_results.to_csv(path.join(output_dir, 'results.tsv'),
                           index=False, sep='\t', encoding='utf-8')

        mean_results = pd.DataFrame(all_results.apply(np.nanmean).to_dict(), columns=all_results.columns, index=[0, ])
        mean_results.to_csv(path.join(output_dir, 'mean_results.tsv'),
                            index=False, sep='\t', encoding='utf-8')

        print("Mean results of the classification:")
        print("Balanced accuracy: %s" % (mean_results['balanced_accuracy'].to_string(index=False)))
        print("specificity: %s" % (mean_results['specificity'].to_string(index=False)))
        print("sensitivity: %s" % (mean_results['sensitivity'].to_string(index=False)))
        print("auc: %s" % (mean_results['auc'].to_string(index=False)))

    @staticmethod
    def get_default_parameters():

        parameters_dict = {'n_folds': 10,
                           'n_threads': 15,
                           'splits_indices': None,
                           'inner_cv': True}

        return parameters_dict


class RepeatedKFoldCV(base.MLValidation):

    def validate(self, y):

        if self._validation_params['splits_indices'] is None:
            self._validation_params['splits_indices'] = []

            for i in range(self._validation_params['n_iterations']):
                skf = StratifiedKFold(n_splits=self._validation_params['n_folds'], shuffle=True)
                self._validation_params['splits_indices'].append(list(skf.split(np.zeros(len(y)), y)))

        async_pool = ThreadPool(self._validation_params['n_threads'])
        async_result = {}

        for i in range(self._validation_params['n_iterations']):

            train_index, test_index = self._validation_params['splits_indices'][i]
            async_result[i] = async_pool.apply_async(self._ml_algorithm.evaluate, (train_index, test_index))

        for r in range(self._validation_params['n_iterations']):

            async_result[r] = {}
            self._validation_results.append([])

            for i in range(self._validation_params['n_folds']):

                train_index, test_index = self._validation_params['splits_indices'][r][i]
                async_result[r][i] = async_pool.apply_async(self._ml_algorithm.evaluate, (train_index, test_index))

        async_pool.close()
        async_pool.join()
        for r in range(self._validation_params['n_iterations']):
            for i in range(self._validation_params['n_folds']):
                self._validation_results[r].append(async_result[r][i].get())

        # TODO Find a better way to estimate best parameter
        flat_results = [result for fold in self._validation_results for result in fold]
        self._classifier, self._best_params = self._ml_algorithm.apply_best_parameters(flat_results)

        return self._classifier, self._best_params, self._validation_results

    def save_results(self, output_dir):
        if self._validation_results is None:
            raise Exception("No results to save. Method validate() must be run before save_results().")

        all_results_list = []
        all_subjects_list = []

        for iteration in range(len(self._validation_results)):

            iteration_dir = path.join(output_dir, 'iteration-' + str(iteration))
            if not path.exists(iteration_dir):
                os.makedirs(iteration_dir)

            iteration_subjects_list = []
            iteration_results_list = []
            folds_dir = path.join(iteration_dir, 'folds')

            if not path.exists(folds_dir):
                os.makedirs(folds_dir)

            for i in range(len(self._validation_results[iteration])):
                subjects_df = pd.DataFrame({'y': self._validation_results[iteration][i]['y'],
                                            'y_hat': self._validation_results[iteration][i]['y_hat'],
                                            'y_index': self._validation_results[iteration][i]['y_index']})
                subjects_df.to_csv(path.join(folds_dir, 'subjects_fold-' + str(i) + '.tsv'),
                                   index=False, sep='\t', encoding='utf-8')
                iteration_subjects_list.append(subjects_df)

                results_df = pd.DataFrame(
                    {'balanced_accuracy': self._validation_results[iteration][i]['evaluation']['balanced_accuracy'],
                     'auc': self._validation_results[iteration][i]['auc'],
                     'accuracy': self._validation_results[iteration][i]['evaluation']['accuracy'],
                     'sensitivity': self._validation_results[iteration][i]['evaluation']['sensitivity'],
                     'specificity': self._validation_results[iteration][i]['evaluation']['specificity'],
                     'ppv': self._validation_results[iteration][i]['evaluation']['ppv'],
                     'npv': self._validation_results[iteration][i]['evaluation']['npv'],
                     'train_balanced_accuracy': self._validation_results[iteration][i]['evaluation_train']['balanced_accuracy'],
                     'train_accuracy': self._validation_results[iteration][i]['evaluation_train']['accuracy'],
                     'train_sensitivity': self._validation_results[iteration][i]['evaluation_train']['sensitivity'],
                     'train_specificity': self._validation_results[iteration][i]['evaluation_train']['specificity'],
                     'train_ppv': self._validation_results[iteration][i]['evaluation_train']['ppv'],
                     'train_npv': self._validation_results[iteration][i]['evaluation_train']['npv']
                     }, index=['i', ])
                results_df.to_csv(path.join(folds_dir, 'results_fold-' + str(i) + '.tsv'),
                                  index=False, sep='\t', encoding='utf-8')
                iteration_results_list.append(results_df)

            iteration_subjects_df = pd.concat(iteration_subjects_list)
            iteration_subjects_df.to_csv(path.join(iteration_dir, 'subjects.tsv'),
                                         index=False, sep='\t', encoding='utf-8')
            all_subjects_list.append(iteration_subjects_df)

            iteration_results_df = pd.concat(iteration_results_list)
            iteration_results_df.to_csv(path.join(iteration_dir, 'results.tsv'),
                                        index=False, sep='\t', encoding='utf-8')

            mean_results_df = pd.DataFrame(iteration_results_df.apply(np.nanmean).to_dict(),
                                           columns=iteration_results_df.columns, index=[0, ])
            mean_results_df.to_csv(path.join(iteration_dir, 'mean_results.tsv'),
                                   index=False, sep='\t', encoding='utf-8')
            all_results_list.append(mean_results_df)

        all_subjects_df = pd.concat(all_subjects_list)
        all_subjects_df.to_csv(path.join(output_dir, 'subjects.tsv'),
                               index=False, sep='\t', encoding='utf-8')

        all_results_df = pd.concat(all_results_list)
        all_results_df.to_csv(path.join(output_dir, 'results.tsv'),
                              index=False, sep='\t', encoding='utf-8')

        mean_results_df = pd.DataFrame(all_results_df.apply(np.nanmean).to_dict(),
                                       columns=all_results_df.columns, index=[0, ])
        mean_results_df.to_csv(path.join(output_dir, 'mean_results.tsv'),
                               index=False, sep='\t', encoding='utf-8')

        print("Mean results of the classification:")
        print("Balanced accuracy: %s" % (mean_results_df['balanced_accuracy'].to_string(index=False)))
        print("specificity: %s" % (mean_results_df['specificity'].to_string(index=False)))
        print("sensitivity: %s" % (mean_results_df['sensitivity'].to_string(index=False)))
        print("auc: %s" % (mean_results_df['auc'].to_string(index=False)))

    @staticmethod
    def get_default_parameters():

        parameters_dict = {'n_iterations': 100,
                           'n_folds': 10,
                           'n_threads': 15,
                           'splits_indices': None,
                           'inner_cv': True}

        return parameters_dict


class RepeatedHoldOut(base.MLValidation):

    def validate(self, y):

        if self._validation_params['splits_indices'] is None:
            splits = StratifiedShuffleSplit(n_splits=self._validation_params['n_iterations'],
                                            test_size=self._validation_params['test_size'])
            self._validation_params['splits_indices'] = list(splits.split(np.zeros(len(y)), y))

        async_pool = ThreadPool(self._validation_params['n_threads'])
        async_result = {}

        for i in range(self._validation_params['n_iterations']):

            train_index, test_index = self._validation_params['splits_indices'][i]
            if self._validation_params['inner_cv']:
                async_result[i] = async_pool.apply_async(self._ml_algorithm.evaluate, (train_index, test_index))
            else:
                async_result[i] = async_pool.apply_async(self._ml_algorithm.evaluate_no_cv, (train_index, test_index))

        async_pool.close()
        async_pool.join()

        for i in range(self._validation_params['n_iterations']):
            self._validation_results.append(async_result[i].get())

        self._classifier, self._best_params = self._ml_algorithm.apply_best_parameters(self._validation_results)
        return self._classifier, self._best_params, self._validation_results

    def save_results(self, output_dir):
        if self._validation_results is None:
            raise Exception("No results to save. Method validate() must be run before save_results().")

        all_results_list = []
        all_train_subjects_list = []
        all_test_subjects_list = []

        for iteration in range(len(self._validation_results)):

            iteration_dir = path.join(output_dir, 'iteration-' + str(iteration))
            if not path.exists(iteration_dir):
                os.makedirs(iteration_dir)
            iteration_train_subjects_df = pd.DataFrame({'iteration': iteration,
                                                        'y': self._validation_results[iteration]['y_train'],
                                                        'y_hat': self._validation_results[iteration]['y_hat_train'],
                                                        'subject_index': self._validation_results[iteration]['x_index']})
            iteration_train_subjects_df.to_csv(path.join(iteration_dir, 'train_subjects.tsv'),
                                               index=False, sep='\t', encoding='utf-8')
            all_train_subjects_list.append(iteration_train_subjects_df)

            iteration_test_subjects_df = pd.DataFrame({'iteration': iteration,
                                                       'y': self._validation_results[iteration]['y'],
                                                       'y_hat': self._validation_results[iteration]['y_hat'],
                                                       'subject_index': self._validation_results[iteration]['y_index']})
            iteration_test_subjects_df.to_csv(path.join(iteration_dir, 'test_subjects.tsv'),
                                              index=False, sep='\t', encoding='utf-8')
            all_test_subjects_list.append(iteration_test_subjects_df)

            iteration_results_df = pd.DataFrame(
                    {'balanced_accuracy': self._validation_results[iteration]['evaluation']['balanced_accuracy'],
                     'auc': self._validation_results[iteration]['auc'],
                     'accuracy': self._validation_results[iteration]['evaluation']['accuracy'],
                     'sensitivity': self._validation_results[iteration]['evaluation']['sensitivity'],
                     'specificity': self._validation_results[iteration]['evaluation']['specificity'],
                     'ppv': self._validation_results[iteration]['evaluation']['ppv'],
                     'npv': self._validation_results[iteration]['evaluation']['npv'],
                     'train_balanced_accuracy': self._validation_results[iteration]['evaluation_train']['balanced_accuracy'],
                     'train_accuracy': self._validation_results[iteration]['evaluation_train']['accuracy'],
                     'train_sensitivity': self._validation_results[iteration]['evaluation_train']['sensitivity'],
                     'train_specificity': self._validation_results[iteration]['evaluation_train']['specificity'],
                     'train_ppv': self._validation_results[iteration]['evaluation_train']['ppv'],
                     'train_npv': self._validation_results[iteration]['evaluation_train']['npv']
                     }, index=['i', ])
            iteration_results_df.to_csv(path.join(iteration_dir, 'results.tsv'),
                                        index=False, sep='\t', encoding='utf-8')

            # mean_results_df = pd.DataFrame(iteration_results_df.apply(np.nanmean).to_dict(),
            #                                columns=iteration_results_df.columns, index=[0, ])
            # mean_results_df.to_csv(path.join(iteration_dir, 'mean_results.tsv'),
            #                        index=False, sep='\t', encoding='utf-8')
            all_results_list.append(iteration_results_df)

        all_train_subjects_df = pd.concat(all_train_subjects_list)
        all_train_subjects_df.to_csv(path.join(output_dir, 'train_subjects.tsv'),
                                     index=False, sep='\t', encoding='utf-8')

        all_test_subjects_df = pd.concat(all_test_subjects_list)
        all_test_subjects_df.to_csv(path.join(output_dir, 'test_subjects.tsv'),
                                    index=False, sep='\t', encoding='utf-8')

        all_results_df = pd.concat(all_results_list)
        all_results_df.to_csv(path.join(output_dir, 'results.tsv'),
                              index=False, sep='\t', encoding='utf-8')

        mean_results_df = pd.DataFrame(all_results_df.apply(np.nanmean).to_dict(),
                                       columns=all_results_df.columns, index=[0, ])
        mean_results_df.to_csv(path.join(output_dir, 'mean_results.tsv'),
                               index=False, sep='\t', encoding='utf-8')

        print("Mean results of the classification:")
        print("Balanced accuracy: %s" % (mean_results_df['balanced_accuracy'].to_string(index=False)))
        print("specificity: %s" % (mean_results_df['specificity'].to_string(index=False)))
        print("sensitivity: %s" % (mean_results_df['sensitivity'].to_string(index=False)))
        print("auc: %s" % (mean_results_df['auc'].to_string(index=False)))

    @staticmethod
    def get_default_parameters():

        parameters_dict = {'n_iterations': 100,
                           'test_size': 0.2,
                           'n_threads': 15,
                           'splits_indices': None,
                           'inner_cv': True}

        return parameters_dict


class LearningCurveRepeatedHoldOut(base.MLValidation):

    def validate(self, y):

        if self._validation_params['splits_indices'] is None:
            splits = StratifiedShuffleSplit(n_splits=self._validation_params['n_iterations'],
                                            test_size=self._validation_params['test_size'])
            self._validation_params['splits_indices'] = list(splits.split(np.zeros(len(y)), y))

        async_pool = ThreadPool(self._validation_params['n_threads'])
        async_result = {}

        for i in range(self._validation_params['n_iterations']):
            train_index, test_index = self._validation_params['splits_indices'][i]
            async_result[i] = {}

            skf = StratifiedKFold(n_splits=self._validation_params['n_learning_points'], shuffle=False)
            inner_cv_splits = list(skf.split(np.zeros(len(y[train_index])), y[train_index]))

            for j in range(self._validation_params['n_learning_points']):
                inner_train_index = np.concatenate([indexes[1] for indexes in
                                                    inner_cv_splits[:j + 1]]).ravel()
                async_result[i][j] = async_pool.apply_async(self._ml_algorithm.evaluate,
                                                            (train_index[inner_train_index], test_index))

        async_pool.close()
        async_pool.join()

        for j in range(self._validation_params['n_learning_points']):
            learning_point_results = []
            for i in range(self._validation_params['n_iterations']):
                learning_point_results.append(async_result[i][j].get())

            self._validation_results.append(learning_point_results)

        self._classifier = []
        self._best_params = []
        for j in range(self._validation_params['n_learning_points']):
            classifier, best_params = self._ml_algorithm.apply_best_parameters(self._validation_results[j])
            self._classifier.append(classifier)
            self._best_params.append(best_params)

        return self._classifier, self._best_params, self._validation_results

    def save_results(self, output_dir):
        if self._validation_results is None:
            raise Exception("No results to save. Method validate() must be run before save_results().")

        for learning_point in range(self._validation_params['n_learning_points']):

            all_results_list = []
            all_subjects_list = []

            learning_point_dir = path.join(output_dir, 'learning_split-' + str(learning_point))

            for iteration in range(self._validation_params['n_iterations']):

                iteration_dir = path.join(learning_point_dir, 'iteration-' + str(iteration))
                if not path.exists(iteration_dir):
                    os.makedirs(iteration_dir)
                iteration_subjects_df = pd.DataFrame({'y': self._validation_results[learning_point][iteration]['y'],
                                                      'y_hat': self._validation_results[learning_point][iteration]['y_hat'],
                                                      'y_index': self._validation_results[learning_point][iteration]['y_index']})
                iteration_subjects_df.to_csv(path.join(iteration_dir, 'subjects.tsv'),
                                             index=False, sep='\t', encoding='utf-8')
                all_subjects_list.append(iteration_subjects_df)

                iteration_results_df = pd.DataFrame(
                        {'balanced_accuracy': self._validation_results[learning_point][iteration]['evaluation']['balanced_accuracy'],
                         'auc': self._validation_results[learning_point][iteration]['auc'],
                         'accuracy': self._validation_results[learning_point][iteration]['evaluation']['accuracy'],
                         'sensitivity': self._validation_results[learning_point][iteration]['evaluation']['sensitivity'],
                         'specificity': self._validation_results[learning_point][iteration]['evaluation']['specificity'],
                         'ppv': self._validation_results[learning_point][iteration]['evaluation']['ppv'],
                         'npv': self._validation_results[learning_point][iteration]['evaluation']['npv'],
                         'train_balanced_accuracy': self._validation_results[learning_point][iteration]['evaluation_train']['balanced_accuracy'],
                         'train_accuracy': self._validation_results[learning_point][iteration]['evaluation_train']['accuracy'],
                         'train_sensitivity': self._validation_results[learning_point][iteration]['evaluation_train']['sensitivity'],
                         'train_specificity': self._validation_results[learning_point][iteration]['evaluation_train']['specificity'],
                         'train_ppv': self._validation_results[learning_point][iteration]['evaluation_train']['ppv'],
                         'train_npv': self._validation_results[learning_point][iteration]['evaluation_train']['npv']}, index=['i', ])

                iteration_results_df.to_csv(path.join(iteration_dir, 'results.tsv'),
                                            index=False, sep='\t', encoding='utf-8')

                mean_results_df = pd.DataFrame(iteration_results_df.apply(np.nanmean).to_dict(),
                                               columns=iteration_results_df.columns, index=[0, ])
                mean_results_df.to_csv(path.join(iteration_dir, 'mean_results.tsv'),
                                       index=False, sep='\t', encoding='utf-8')
                all_results_list.append(mean_results_df)

            all_subjects_df = pd.concat(all_subjects_list)
            all_subjects_df.to_csv(path.join(learning_point_dir, 'subjects.tsv'),
                                   index=False, sep='\t', encoding='utf-8')

            all_results_df = pd.concat(all_results_list)
            all_results_df.to_csv(path.join(learning_point_dir, 'results.tsv'),
                                  index=False, sep='\t', encoding='utf-8')

            mean_results_df = pd.DataFrame(all_results_df.apply(np.nanmean).to_dict(),
                                           columns=all_results_df.columns, index=[0, ])
            mean_results_df.to_csv(path.join(learning_point_dir, 'mean_results.tsv'),
                                   index=False, sep='\t', encoding='utf-8')

    @staticmethod
    def get_default_parameters():

        parameters_dict = {'n_iterations': 100,
                           'test_size': 0.2,
                           'n_learning_points': 10,
                           'n_threads': 15,
                           'splits_indices': None,
                           'inner_cv': True}

        return parameters_dict


class RepeatedKFoldCV_Multiclass(base.MLValidation):

    def __init__(self, ml_algorithm):
        self._ml_algorithm = ml_algorithm
        self._repeated_validation_results = []
        self._classifier = None
        self._best_params = None
        self._cv = None

    def validate(self, y, n_iterations=100, n_folds=10, n_threads=15):

        async_pool = ThreadPool(self._validation_params['n_threads'])
        async_result = {}
        self._cv = []

        for r in range(n_iterations):
            skf = StratifiedKFold(n_splits=n_folds, shuffle=True)
            self._cv.append(list(skf.split(np.zeros(len(y)), y)))
            async_result[r] = {}
            self._repeated_validation_results.append([])

            for i in range(n_folds):

                train_index, test_index = self._cv[r][i]
                async_result[r][i] = async_pool.apply_async(self._ml_algorithm.evaluate, (train_index, test_index))

        async_pool.close()
        async_pool.join()
        for r in range(n_iterations):
            for i in range(n_folds):
                self._repeated_validation_results[r].append(async_result[r][i].get())

        # TODO Find a better way to estimate best parameter
        flat_results = [result for fold in self._repeated_validation_results for result in fold]
        self._classifier, self._best_params = self._ml_algorithm.apply_best_parameters(flat_results)

        return self._classifier, self._best_params, self._repeated_validation_results

    def save_results(self, output_dir):
        if self._repeated_validation_results is None:
            raise Exception("No results to save. Method validate() must be run before save_results().")

        all_results_list = []
        all_subjects_list = []

        for iteration in range(len(self._repeated_validation_results)):

            iteration_dir = path.join(output_dir, 'iteration-' + str(iteration))
            if not path.exists(iteration_dir):
                os.makedirs(iteration_dir)

            iteration_subjects_list = []
            iteration_results_list = []
            folds_dir = path.join(iteration_dir, 'folds')

            if not path.exists(folds_dir):
                os.makedirs(folds_dir)

            for i in range(len(self._repeated_validation_results[iteration])):
                subjects_df = pd.DataFrame({'y': self._repeated_validation_results[iteration][i]['y'],
                                            'y_hat': self._repeated_validation_results[iteration][i]['y_hat'],
                                            'y_index': self._repeated_validation_results[iteration][i]['y_index']})
                subjects_df.to_csv(path.join(folds_dir, 'subjects_fold-' + str(i) + '.tsv'),
                                   index=False, sep='\t', encoding='utf-8')
                iteration_subjects_list.append(subjects_df)

                results_df = pd.DataFrame(
                    {'balanced_accuracy': self._repeated_validation_results[iteration][i]['evaluation']['balanced_accuracy'],
                     'accuracy': self._repeated_validation_results[iteration][i]['evaluation']['accuracy'],
                     'train_balanced_accuracy': self._repeated_validation_results[iteration][i]['evaluation_train']['balanced_accuracy'],
                     'train_accuracy': self._repeated_validation_results[iteration][i]['evaluation_train']['accuracy']
                     }, index=['i', ])
                results_df.to_csv(path.join(folds_dir, 'results_fold-' + str(i) + '.tsv'),
                                  index=False, sep='\t', encoding='utf-8')
                iteration_results_list.append(results_df)

            iteration_subjects_df = pd.concat(iteration_subjects_list)
            iteration_subjects_df.to_csv(path.join(iteration_dir, 'subjects.tsv'),
                                         index=False, sep='\t', encoding='utf-8')
            all_subjects_list.append(iteration_subjects_df)

            iteration_results_df = pd.concat(iteration_results_list)
            iteration_results_df.to_csv(path.join(iteration_dir, 'results.tsv'),
                                        index=False, sep='\t', encoding='utf-8')

            mean_results_df = pd.DataFrame(iteration_results_df.apply(np.nanmean).to_dict(),
                                           columns=iteration_results_df.columns, index=[0, ])
            mean_results_df.to_csv(path.join(iteration_dir, 'mean_results.tsv'),
                                   index=False, sep='\t', encoding='utf-8')
            all_results_list.append(mean_results_df)

        all_subjects_df = pd.concat(all_subjects_list)
        all_subjects_df.to_csv(path.join(output_dir, 'subjects.tsv'),
                               index=False, sep='\t', encoding='utf-8')

        all_results_df = pd.concat(all_results_list)
        all_results_df.to_csv(path.join(output_dir, 'results.tsv'),
                              index=False, sep='\t', encoding='utf-8')

        mean_results_df = pd.DataFrame(all_results_df.apply(np.nanmean).to_dict(),
                                       columns=all_results_df.columns, index=[0, ])
        mean_results_df.to_csv(path.join(output_dir, 'mean_results.tsv'),
                               index=False, sep='\t', encoding='utf-8')

        print("Mean results of the classification:")
        print("Balanced accuracy: %s" % (mean_results_df['balanced_accuracy'].to_string(index=False)))



# coding: utf8

import clinica.pipelines.engine as cpe


class StatisticsSurface(cpe.Pipeline):
    """
    StatisticsSurface - Surface-based mass-univariate analysis with SurfStat

    See documentation at http://www.clinica.run/doc/Pipelines/Stats_Surface/

    Note:
        The `tsv_file` attribute is overloaded for this pipeline. It must contain a list of subjects
        with their sessions and all the covariates and factors needed for the GLM.

        Pipeline parameters are explained in StatisticsSurfaceCLI.define_options()

    Returns:
        A clinica pipeline object containing the StatisticsSurface pipeline.
    """
    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        from .statistics_surface_utils import get_t1_freesurfer_custom_file
        from clinica.utils.exceptions import ClinicaException
        from clinica.utils.group import check_group_label

        if 'orig_input_data' not in self.parameters.keys():
            raise KeyError('Missing compulsory orig_input_data key in pipeline parameter.')
        if 'covariates' not in self.parameters.keys():
            self.parameters['covariates'] = None
        if 'custom_file' not in self.parameters.keys():
            self.parameters['custom_file'] = get_t1_freesurfer_custom_file()
        if 'measure_label' not in self.parameters.keys():
            self.parameters['measure_label'] = 'ct',
        if 'full_width_at_half_maximum' not in self.parameters.keys():
            self.parameters['full_width_at_half_maximum'] = 20
        if 'cluster_threshold' not in self.parameters.keys():
            self.parameters['cluster_threshold'] = 0.001,

        check_group_label(self.parameters['group_label'])
        if self.parameters['glm_type'] not in ['group_comparison', 'correlation']:
            raise ClinicaException("The glm_type you specified is wrong: it should be group_comparison or "
                                   "correlation (given value: %s)." % self.parameters['glm_type'])
        if self.parameters['full_width_at_half_maximum'] not in [0, 5, 10, 15, 20]:
            raise ClinicaException(
                "FWHM for the surface smoothing you specified is wrong: it should be 0, 5, 10, 15 or 20 "
                "(given value: %s)." % self.parameters['full_width_at_half_maximum'])
        if self.parameters['cluster_threshold'] < 0 or self.parameters['cluster_threshold'] > 1:
            raise ClinicaException("Cluster threshold should be between 0 and 1 "
                                   "(given value: %s)." % self.parameters['cluster_threshold'])

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file.
        """
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipelines.

        Returns:
            A list of (string) input fields name.
        """
        return []

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipelines.

        Returns:
            A list of (string) output fields name.
        """
        return ['output_dir']

    def build_input_node(self):
        """Build and connect an input node to the pipelines.
        """
        import os
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.exceptions import ClinicaException
        from clinica.utils.stream import cprint

        # Check if already present in CAPS
        # ================================
        # Check if the group label has been existed, if yes, give an error to the users
        # Note(AR): if the user wants to compare Cortical Thickness measure with PET measure
        # using the group_id, Clinica won't allow it.
        # TODO: Modify this behaviour
        if os.path.exists(os.path.join(self.caps_directory, 'groups', 'group-' + self.parameters['group_label'])):
            error_message = ('Group ID %s already exists, please choose another one or delete the existing folder and '
                             'also the working directory and rerun the pipeline') % self.parameters['group_label']
            raise ClinicaException(error_message)
            # statistics_dir_tsv = os.path.join(input_directory, 'groups', group_id, 'statistics', 'participant.tsv')
            # # Copy the subjects_visits_tsv to the result folder
            # # First, check if the subjects_visits_tsv has the same info with the participant.tsv in the folder of statistics.
            # # If the participant TSV does not exit, copy subjects_visits_tsv in the folder of statistics too,
            # # if it is here, compare them.
            # if not os.path.isfile(statistics_dir_tsv):
            #     copy(subjects_visits_tsv, statistics_dir_tsv)
            # else:
            #     # Compare the two TSV files
            #     if not have_same_subjects(statistics_dir_tsv, subjects_visits_tsv):
            #         raise ValueError("It seems that this round of analysis does not contain the same subjects"
            #                          "where you want to put the results, please check it!")

        # Check input files before calling SurfStat with Matlab
        # =====================================================
        all_errors = []
        # clinica_files_reader expects regexp to start at subjects/ so sub-*/ses-*/ is removed here
        pattern_hemisphere = self.parameters['custom_file'].replace(
            '@subject', 'sub-*').replace(
            '@session', 'ses-*').replace(
            '@fwhm', str(self.parameters['full_width_at_half_maximum'])).replace(
            'sub-*/ses-*/', ''
        )
        # Files on left hemisphere
        lh_surface_based_info = {
            'pattern': pattern_hemisphere.replace('@hemi', 'lh'),
            'description': 'surface-based features on left hemisphere at FWHM = %s' %
                           self.parameters['full_width_at_half_maximum'],
        }
        try:
            clinica_file_reader(self.subjects, self.sessions, self.caps_directory, lh_surface_based_info)
        except ClinicaException as e:
            all_errors.append(e)
        rh_surface_based_info = {
            'pattern': pattern_hemisphere.replace('@hemi', 'rh'),
            'description': 'surface-based features on right hemisphere at FWHM = %s' %
                           self.parameters['full_width_at_half_maximum'],
        }
        try:
            clinica_file_reader(self.subjects, self.sessions, self.caps_directory, rh_surface_based_info)
        except ClinicaException as e:
            all_errors.append(e)
        # Raise all errors if something happened
        if len(all_errors) > 0:
            error_message = 'Clinica faced errors while trying to read files in your CAPS directory.\n'
            for msg in all_errors:
                error_message += str(msg)
            raise RuntimeError(error_message)

        # Give pipeline info
        # ==================
        cprint('The pipeline will last a few minutes. Images generated by Matlab will popup during the pipeline.')

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from .statistics_surface_utils import save_to_caps

        # Writing results into CAPS
        # =========================
        save_to_caps = npe.Node(interface=nutil.Function(
            input_names=['source_dir', 'caps_dir', 'overwrite_caps', 'pipeline_parameters'],
            function=save_to_caps),
            name='SaveToCaps')
        save_to_caps.inputs.caps_dir = self.caps_directory
        save_to_caps.inputs.overwrite_caps = self.overwrite_caps
        save_to_caps.inputs.pipeline_parameters = self.parameters

        self.connect([
            (self.output_node, save_to_caps, [('output_dir', 'source_dir')]),
        ])

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import os
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from .statistics_surface_utils import init_input_node
        import clinica.pipelines.statistics_surface.statistics_surface_utils as utils

        init_input = npe.Node(
            interface=nutil.Function(
                input_names=['parameters', 'base_dir', 'subjects_visits_tsv'],
                output_names=['group_label', 'surfstat_results_dir'],
                function=init_input_node),
            name='0-InitPipeline')
        init_input.inputs.parameters = self.parameters
        init_input.inputs.base_dir = os.path.join(self.base_dir, self.name)
        init_input.inputs.subjects_visits_tsv = self.tsv_file

        # Node to wrap the SurfStat matlab script
        surfstat = npe.Node(name='1-RunSurfStat',
                            interface=nutil.Function(
                                input_names=['caps_dir',
                                             'output_dir',
                                             'subjects_visits_tsv',
                                             'pipeline_parameters',
                                             ],
                                output_names=['output_dir'],
                                function=utils.run_matlab))
        surfstat.inputs.caps_dir = self.caps_directory
        surfstat.inputs.subjects_visits_tsv = self.tsv_file
        surfstat.inputs.pipeline_parameters = self.parameters

        # Connection
        # ==========
        self.connect([
            (init_input, surfstat, [('surfstat_results_dir', 'output_dir')]),
            (surfstat, self.output_node, [('output_dir', 'output_dir')]),
        ])



# coding: utf8

import clinica.engine as ce


class T1VolumeParcellationCLI(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipeline."""
        self._name = 't1-volume-parcellation'

    def define_description(self):
        """Define a description of this pipeline."""
        self._description = ('Computation of mean GM concentration for a set of regions:\n'
                             'http://clinica.run/doc/Pipelines/T1_Volume/')

    def define_options(self):
        """Define the sub-command arguments."""
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES
        # Clinica compulsory arguments (e.g. BIDS, CAPS, group_id)
        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')
        clinica_comp.add_argument("group_id",
                                  help='User-defined identifier for the provided group of subjects.')
        # Clinica standard arguments (e.g. --n_procs)
        self.add_clinica_standard_arguments()

    def run_command(self, args):
        """Run the pipeline with defined args."""
        from networkx import Graph
        from .t1_volume_parcellation_pipeline import T1VolumeParcellation
        from clinica.utils.ux import print_end_pipeline, print_crash_files_and_exit

        parameters = {
            'group_id': args.group_id,
        }
        pipeline = T1VolumeParcellation(
            caps_directory=self.absolute_path(args.caps_directory),
            tsv_file=self.absolute_path(args.subjects_sessions_tsv),
            base_dir=self.absolute_path(args.working_directory),
            parameters=parameters,
            name=self.name
        )

        if args.n_procs:
            exec_pipeline = pipeline.run(plugin='MultiProc',
                                         plugin_args={'n_procs': args.n_procs})
        else:
            exec_pipeline = pipeline.run()

        if isinstance(exec_pipeline, Graph):
            print_end_pipeline(self.name, pipeline.base_dir, pipeline.base_dir_was_specified)
        else:
            print_crash_files_and_exit(args.logname, pipeline.base_dir)



# coding: utf8

import clinica.engine as ce


class AiblToBidsCLI(ce.CmdParser):
    def define_name(self):
        """Define the sub-command name to run this command."""
        self._name = 'aibl-to-bids'

    def define_description(self):
        """Define a description of this command."""
        self._description = 'Convert AIBL (https://aibl.csiro.au/adni/index.html) into BIDS.'

    def define_options(self):
        """Define the sub-command arguments."""
        self._args.add_argument("dataset_directory",
                                help='Path to the AIBL images directory.')
        self._args.add_argument("clinical_data_directory",
                                help='Path to the AIBL clinical data directory.')
        self._args.add_argument("bids_directory",
                                help='Path to the BIDS directory.')

    def run_command(self, args):
        """Run the converter with defined args."""
        from os.path import exists
        from os import makedirs
        from clinica.iotools.converters.aibl_to_bids.aibl_to_bids import convert_clinical_data, convert_images
        from clinica.utils.check_dependency import check_dcm2nii, check_dcm2niix, check_freesurfer

        check_dcm2nii()
        check_dcm2niix()
        check_freesurfer()

        if not exists(args.bids_directory):
            makedirs(args.bids_directory)

        convert_images(args.dataset_directory, args.clinical_data_directory, args.bids_directory)
        convert_clinical_data(args.bids_directory, args.clinical_data_directory)



# coding: utf8

def extract_slices(input_tensor, slice_direction=0, slice_mode='single'):
    """Extracts the slices from three directions
    
    This function extracts slices form the preprocesed nifti image.  The
    direction of extraction can be defined either on sagital direction (0),
    cornal direction (1) or axial direction (other). The output slices can be
    stores following two modes: single (1 channel) ou RGB (3 channels, all the
    same).

    
    Args:
        input_tensor: tensor version of the nifti MRI.
        slice_direction: which axis direction that the slices were extracted
        slice_mode: 'single' or 'RGB'.

    Returns:
        file: multiple tensors saved on the disk, suffixes corresponds to
            indexes of the slices. Same location than input file.
    """
    import torch
    import os

    image_tensor = torch.load(input_tensor)
    # reshape the tensor, delete the first dimension for slice-level
    image_tensor = image_tensor.view(image_tensor.shape[1], image_tensor.shape[2], image_tensor.shape[3])

    # sagital
    # M and N correspond to the first and last slices (if need to remove)
    M = 0
    N = 0
    slice_list_sag = range(M, image_tensor.shape[0] - N)  # delete the first M slices and last N slices

    basedir = os.getcwd()
    input_tensor_filename = os.path.basename(input_tensor)
    
    txt_idx = input_tensor_filename.rfind("_")
    it_filename_prefix = input_tensor_filename[0:txt_idx]
    it_filename_suffix = input_tensor_filename[txt_idx:]
    
    output_file_original = []
    output_file_rgb = []
    if slice_direction == 0:
        for index_slice, index_slice_list in zip(slice_list_sag, range(len(slice_list_sag))):
            # for i in slice_list:
            # sagital
            slice_select_sag = image_tensor[index_slice, :, :]

            extracted_slice_original_sag = slice_select_sag.unsqueeze(0)  # shape should be 1 * W * L

            # train for transfer learning, creating the fake RGB image.
            slice_select_sag = (slice_select_sag - slice_select_sag.min()) / (slice_select_sag.max() - slice_select_sag.min())
            extracted_slice_rgb_sag = torch.stack((slice_select_sag, slice_select_sag, slice_select_sag))  # shape should be 3 * W * L

            # save into .pt format
            if slice_mode == 'single':
                output_file_original.append(
                        os.path.join(
                            basedir,
                            it_filename_prefix
                            + '_axis-sag_channel-single_slice-'
                            + str(index_slice)
                            + it_filename_suffix
                            )
                        )
                torch.save(extracted_slice_original_sag.clone(), output_file_original[index_slice_list])
            elif slice_mode == 'rgb':
                output_file_rgb.append(
                        os.path.join(
                            basedir,
                            it_filename_prefix
                            + '_axis-sag_channel-rgb_slice-'
                            + str(index_slice)
                            + it_filename_suffix
                            )
                        )
                torch.save(extracted_slice_rgb_sag.clone(), output_file_rgb[index_slice_list])

    elif slice_direction == 1:
        # cornal
        slice_list_cor = range(M, image_tensor.shape[1] - N)  # delete the first M slices and last N slices
        for index_slice, index_slice_list in zip(slice_list_cor, range(len(slice_list_cor))):
            # for i in slice_list:
            # sagital
            slice_select_cor = image_tensor[:, index_slice, :]

            extracted_slice_original_cor = slice_select_cor.unsqueeze(0)  # shape should be 1 * W * L

            # train for transfer learning, creating the fake RGB image.
            slice_select_cor = (slice_select_cor - slice_select_cor.min()) / (slice_select_cor.max() - slice_select_cor.min())
            extracted_slice_rgb_cor = torch.stack((slice_select_cor, slice_select_cor, slice_select_cor))  # shape should be 3 * W * L

            # save into .pt format
            if slice_mode == 'single':
                output_file_original.append(
                        os.path.join(
                            basedir,
                            it_filename_prefix
                            + '_axis-cor_channel-single_slice-'
                            + str(index_slice)
                            + it_filename_suffix
                            )
                        )
                torch.save(extracted_slice_original_cor.clone(), output_file_original[index_slice_list])
            elif slice_mode == 'rgb':
                output_file_rgb.append(
                    os.path.join(
                        basedir,
                        it_filename_prefix
                        + '_axis-cor_channel-rgb_slice-'
                        + str(index_slice)
                        + it_filename_suffix
                        )
                    )
                torch.save(extracted_slice_rgb_cor.clone(), output_file_rgb[index_slice_list])

    else:

        # axial
        slice_list_axi = range(M, image_tensor.shape[2] - N)  # delete the first M slices and last N slices
        for index_slice, index_slice_list in zip(slice_list_axi, range(len(slice_list_axi))):
            # for i in slice_list:
            # sagital
            slice_select_axi = image_tensor[:, :, index_slice]

            extracted_slice_original_axi = slice_select_axi.unsqueeze(0)  # shape should be 1 * W * L

            # train for transfer learning, creating the fake RGB image.
            slice_select_axi = (slice_select_axi - slice_select_axi.min()) / (slice_select_axi.max() - slice_select_axi.min())
            extracted_slice_rgb_axi = torch.stack((slice_select_axi, slice_select_axi, slice_select_axi))  # shape should be 3 * W * L

            # save into .pt format
            if slice_mode == 'single':
                output_file_original.append(
                        os.path.join(
                            basedir,
                            it_filename_prefix
                            + '_axis-axi_channel-single_slice-'
                            + str(index_slice)
                            + it_filename_suffix
                            )
                        )
                torch.save(extracted_slice_original_axi.clone(), output_file_original[index_slice_list])
            elif slice_mode == 'rgb':
                output_file_rgb.append(
                        os.path.join(
                            basedir,
                            it_filename_prefix
                            + '_axis-axi_channel-rgb_slice-'
                            + str(index_slice)
                            + it_filename_suffix
                            )
                        )
                torch.save(extracted_slice_rgb_axi.clone(), output_file_rgb[index_slice_list])

    return output_file_rgb, output_file_original


def extract_patches(input_tensor, patch_size, stride_size):
    """Extracts the patches
    
    This function extracts patches form the preprocesed nifti image. Patch size
    if provieded as input and also the stride size. If stride size is smaller
    than the patch size an overlap exist between consecutive patches. If stride
    size is equal to path size there is no overlap. Otherwise, unprocessed
    zones can exits.
    
    Args:
        input_tensor: tensor version of the nifti MRI.
        patch_size: size of a single patch.
        stride_size: size of the stride leading to next patch.

    Returns:
        file: multiple tensors saved on the disk, suffixes corresponds to
            indexes of the patches. Same location than input file.
    """
    import torch
    import os

    basedir = os.getcwd()
    image_tensor = torch.load(input_tensor)

    # use classifiers tensor.upfold to crop the patch.
    patches_tensor = image_tensor.unfold(1, patch_size, stride_size).unfold(2, patch_size, stride_size).unfold(3, patch_size, stride_size).contiguous()
    # the dimension of patch_tensor should be [1, patch_num1, patch_num2, patch_num3, patch_size1, patch_size2, patch_size3]
    patches_tensor = patches_tensor.view(-1, patch_size, patch_size, patch_size)
    
    input_tensor_filename = os.path.basename(input_tensor)
    txt_idx = input_tensor_filename.rfind("_")
    it_filename_prefix = input_tensor_filename[0:txt_idx]
    it_filename_suffix = input_tensor_filename[txt_idx:]

    output_patch = []
    for index_patch in range(patches_tensor.shape[0]):
        extracted_patch = patches_tensor[index_patch, ...].unsqueeze_(0)  # add one dimension
        # save into .pt format
        output_patch.append(
                os.path.join(
                    basedir,
                    it_filename_prefix
                    + '_patchsize-'
                    + str(patch_size)
                    + '_stride-'
                    + str(stride_size)
                    + '_patch-'
                    + str(index_patch)
                    + it_filename_suffix
                    )
                )
        torch.save(extracted_patch.clone(), output_patch[index_patch])

    return output_patch


def save_as_pt(input_img):
    """Saves PyTorch tensor version of the nifti image
    
    This function convert nifti image to tensor (.pt) version of the image.
    Tensor version is saved at the same location than input_img.
    
    Args:
        input_tensor: tensor version of the nifti MRI.

    Returns:
        filename (str): single tensor file  saved on the disk. Same location than input file.

    """

    import torch
    import os
    import nibabel as nib

    basedir = os.getcwd()
    image_array = nib.load(input_img).get_fdata()
    image_tensor = torch.from_numpy(image_array).unsqueeze(0).float()
    # make sure the tensor dtype is torch.float32
    output_file = os.path.join(basedir, os.path.basename(input_img).split('.nii.gz')[0] + '.pt')
    # save
    torch.save(image_tensor.clone(), output_file)

    return output_file



# coding: utf-8

"""
Module for converting DWI of ADNI
"""


def convert_adni_dwi(source_dir, csv_dir, dest_dir, subjs_list=None):
    """
    Convert DW images of ADNI into BIDS format

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    """

    import pandas as pd
    from os import path
    from clinica.utils.stream import cprint
    from colorama import Fore
    from clinica.iotools.converters.adni_to_bids.adni_utils import paths_to_bids

    if subjs_list is None:
        adni_merge_path = path.join(csv_dir, 'ADNIMERGE.csv')
        adni_merge = pd.read_csv(adni_merge_path, sep=',', low_memory=False)
        subjs_list = list(adni_merge.PTID.unique())

    cprint('Calculating paths of DWI images. Output will be stored in ' + path.join(dest_dir, 'conversion_info') + '.')
    images = compute_dwi_paths(source_dir, csv_dir, dest_dir, subjs_list)
    cprint('Paths of DWI images found. Exporting images into BIDS ...')
    # dwi_paths_to_bids(images, dest_dir)
    paths_to_bids(images, dest_dir, 'dwi')
    cprint(Fore.GREEN + 'DWI conversion done.' + Fore.RESET)


def compute_dwi_paths(source_dir, csv_dir, dest_dir, subjs_list):
    """
    Compute paths to DW images to convert to BIDS

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    Returns:
        images: pandas dataframe that contains the path to all the DW images to convert

    """

    from os import path, mkdir
    import pandas as pd

    from clinica.iotools.converters.adni_to_bids.adni_utils import find_image_path, visits_to_timepoints

    dwi_col_df = ['Subject_ID', 'VISCODE', 'Visit', 'Sequence', 'Scan_Date',
                  'Study_ID', 'Series_ID', 'Image_ID', 'Field_Strength']
    dwi_df = pd.DataFrame(columns=dwi_col_df)
    dwi_dfs_list = []

    # Loading needed .csv files
    adni_merge = pd.read_csv(path.join(csv_dir, 'ADNIMERGE.csv'), sep=',', low_memory=False)

    mayo_mri_qc = pd.read_csv(path.join(csv_dir, 'MAYOADIRL_MRI_IMAGEQC_12_08_15.csv'), sep=',', low_memory=False)
    mayo_mri_qc = mayo_mri_qc[mayo_mri_qc.series_type == 'DTI']

    mri_list = pd.read_csv(path.join(csv_dir, 'MRILIST.csv'), sep=',', low_memory=False)

    # Selecting only DTI images that are not Multiband, processed or enhanced images
    mri_list = mri_list[mri_list.SEQUENCE.str.contains('dti', case=False, na=False)]
    unwanted_sequences = ['MB', 'ADC', 'FA', 'TRACEW', 'Enhanced', 'Reg']
    mri_list = mri_list[mri_list.SEQUENCE.map(lambda x: not any(subs in x for subs in unwanted_sequences))]

    for subj in subjs_list:

        # Filter ADNIMERGE, MRI_LIST and QC for only one subject and sort the rows/visits by examination date
        adnimerge_subj = adni_merge[adni_merge.PTID == subj]
        adnimerge_subj = adnimerge_subj.sort_values('EXAMDATE')

        mri_list_subj = mri_list[mri_list.SUBJECT == subj]
        mri_list_subj = mri_list_subj.sort_values('SCANDATE')

        mayo_mri_qc_subj = mayo_mri_qc[mayo_mri_qc.RID == int(subj[-4:])]

        # Obtain corresponding timepoints for the subject visits
        visits = visits_to_timepoints(subj, mri_list_subj, adnimerge_subj, "DWI")

        for visit_info in visits.keys():
            timepoint = visit_info[0]
            visit_str = visits[visit_info]

            visit_mri_list = mri_list_subj[mri_list_subj.VISIT == visit_str]
            axial = dwi_image(subj, timepoint, visits[visit_info], visit_mri_list, mayo_mri_qc_subj)

            if axial is not None:
                row_to_append = pd.DataFrame(axial, index=['i', ])
                dwi_dfs_list.append(row_to_append)

    if dwi_dfs_list:
        dwi_df = pd.concat(dwi_dfs_list, ignore_index=True)

    # Exceptions
    # ==========
    conversion_errors = [('029_S_2395', 'm60'),
                         ('029_S_0824', 'm108'),
                         ('029_S_0914', 'm108'),
                         ('027_S_2219', 'm36'),
                         ('129_S_2332', 'm12'),
                         ('029_S_4384', 'm48'),
                         ('029_S_4385', 'm48'),
                         ('029_S_4585', 'm48'),
                         ('016_S_4591', 'm24'),
                         ('094_S_4630', 'm06'),
                         ('094_S_4649', 'm06'),
                         ('029_S_5219', 'm24'),
                         ('094_S_2238', 'm48'),
                         ('129_S_4287', 'bl'),
                         ('007_S_4611', 'm03'),
                         ('016_S_4638', 'bl'),
                         ('027_S_5118', 'bl'),
                         ('098_S_4018', 'bl'),
                         ('098_S_4003', 'm12'),
                         ('016_S_4584', 'm24'),
                         ('016_S_5007', 'm12'),
                         ('129_S_2347', 'm06'),
                         ('129_S_4220', 'bl'),
                         ('007_S_2058', 'm12'),
                         ('016_S_2007', 'm06'),
                         ('020_S_6358', 'bl'),
                         ('114_S_6039', 'm12'),
                         ('114_S_6057', 'bl'),
                         ('153_S_6274', 'bl'),
                         ('006_S_4485', 'm84'),
                         ('153_S_6237', 'bl'),
                         ('153_S_6336', 'bl'),
                         ('153_S_6450', 'bl'),
                         ('003_S_4441', 'm12'),
                         # Several output images
                         ('029_S_2395', 'm72')]

    # Removing known exceptions from images to convert
    if not dwi_df.empty:
        error_ind = dwi_df.index[dwi_df.apply(lambda x: ((x.Subject_ID, x.VISCODE) in conversion_errors), axis=1)]
        dwi_df.drop(error_ind, inplace=True)

    # Checking for images paths in filesystem
    images = find_image_path(dwi_df, source_dir, 'DWI', 'S', 'Series_ID')

    dwi_tsv_path = path.join(dest_dir, 'conversion_info')
    if not path.exists(dwi_tsv_path):
        mkdir(dwi_tsv_path)
    images.to_csv(path.join(dwi_tsv_path, 'dwi_paths.tsv'), sep='\t', index=False)

    return images


def dwi_image(subject_id, timepoint, visit_str, visit_mri_list, mri_qc_subj):
    """
    One image among those in the input list is chosen according to QC
    and then correspoding metadata is extracted to a dictionary

    Args:
        subject_id: Subject identifier
        timepoint: Visit code
        visit_str: Visit name
        visit_mri_list: List of images metadata
        mri_qc_subj: Dataframe containing list of QC of scans for the subject

    Returns: dictionary - contains image metadata

    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import replace_sequence_chars, select_image_qc

    sel_image = select_image_qc(list(visit_mri_list.IMAGEUID), mri_qc_subj)
    if sel_image is None:
        return None

    sel_scan = visit_mri_list[visit_mri_list.IMAGEUID == sel_image].iloc[0]

    image_dict = {'Subject_ID': subject_id,
                  'VISCODE': timepoint,
                  'Visit': visit_str,
                  'Sequence': replace_sequence_chars(sel_scan.SEQUENCE),
                  'Scan_Date': sel_scan['SCANDATE'],
                  'Study_ID': str(int(sel_scan.STUDYID)),
                  'Series_ID': str(int(sel_scan.SERIESID)),
                  'Image_ID': str(int(sel_scan.IMAGEUID)),
                  'Field_Strength': sel_scan.MAGSTRENGTH}

    return image_dict



# coding: utf8


def get_luts():
    import os
    from clinica.utils.exceptions import ClinicaException

    try:
        # For aparc+aseg.mgz file:
        default = os.path.join(os.environ['FREESURFER_HOME'],
                               'FreeSurferColorLUT.txt')
        # For aparc.a2009s+aseg.mgz file:
        a2009s = os.path.join(os.environ['FREESURFER_HOME'],
                              'FreeSurferColorLUT.txt')

        # TODO: Add custom Lausanne2008 LUTs here.
    except KeyError:
        raise ClinicaException('Could not find FREESURFER_HOME environment variable.')

    return [default, a2009s]


def get_conversion_luts():
    import os
    from clinica.utils.exceptions import ClinicaException

    try:
        # For aparc+aseg.mgz file:
        default = os.path.join(os.environ['MRTRIX_HOME'],
                               'share/mrtrix3/labelconvert/fs_default.txt')
        # For aparc.a2009s+aseg.mgz file:
        a2009s = os.path.join(os.environ['MRTRIX_HOME'],
                              'share/mrtrix3/labelconvert/fs_a2009s.txt')

        # TODO: Add custom Lausanne2008 conversion LUTs here.
    except KeyError:
        raise ClinicaException('Could not find MRTRIX_HOME environment variable.')

    return [default, a2009s]


def get_containers(subjects, sessions):

    return [
        'subjects/' + subjects[i] + '/' + sessions[i] + '/dwi'
        for i in range(len(subjects))
    ]


def get_caps_filenames(dwi_file):

    import re

    m = re.search(r'\/(sub-[a-zA-Z0-9]+_ses-[a-zA-Z0-9]+.*)_preproc', dwi_file)
    if m is None:
        raise ValueError('Input filename is not in a CAPS compliant format.')
    source_file_caps = m.group(1)

    m = re.search(r'\/(sub-[a-zA-Z0-9]+_ses-[a-zA-Z0-9]+.*)_space-[a-zA-Z0-9]+_preproc', dwi_file)
    if m is None:
        raise ValueError('Input filename is not in a CAPS compliant format.')
    source_file_bids = m.group(1)

    response = source_file_caps + '_model-CSD_responseFunction.txt'
    fod = source_file_caps + '_model-CSD_diffmodel.nii.gz'
    tracts = source_file_caps + '_model-CSD_tractography.tck'
    nodes = [source_file_caps + '_atlas-desikan_parcellation.nii.gz',
             source_file_caps + '_atlas-destrieux_parcellation.nii.gz']
    # TODO: Add custom Lausanne2008 node files here.
    connectomes = [source_file_bids + '_model-CSD_atlas-desikan_connectivity.tsv',
                   source_file_bids + '_model-CSD_atlas-destrieux_connectivity.tsv']
    # TODO: Add custom Lausanne2008 connectome files here.

    return response, fod, tracts, nodes, connectomes


def print_begin_pipeline(in_bids_or_caps_file):
    """
    """
    from clinica.utils.stream import cprint
    import re
    import datetime
    from colorama import Fore

    m = re.search(r'(sub-[a-zA-Z0-9]+)_(ses-[a-zA-Z0-9]+)',
                  in_bids_or_caps_file)
    if m is None:
        raise ValueError(
            'Input filename is not in a BIDS or CAPS compliant format.')
    now = datetime.datetime.now().strftime('%H:%M:%S')

    cprint('%s[%s]%s Running pipeline for %s...' % (
        Fore.BLUE, now, Fore.RESET, m.group(0)))


def print_end_pipeline(in_bids_or_caps_file, final_file):
    """
    """
    from clinica.utils.stream import cprint
    import re
    import datetime
    from colorama import Fore

    m = re.search(r'(sub-[a-zA-Z0-9]+)_(ses-[a-zA-Z0-9]+)',
                  in_bids_or_caps_file)
    if m is None:
        raise ValueError(
            'Input filename is not in a BIDS or CAPS compliant format.')
    now = datetime.datetime.now().strftime('%H:%M:%S')

    cprint('%s[%s]%s ...%s has completed.' % (
        Fore.GREEN, now, Fore.RESET, m.group(0)))



# coding: utf-8

import clinica.engine as ce


class StatisticsVolumeCLI(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipeline."""
        self._name = 'statistics-volume'

    def define_description(self):
        """Define a description of this pipeline."""
        self._description = ('Volume-based mass-univariate analysis with SPM:\n'
                             'http://clinica.run/doc/Pipelines/Statistics_Volume/')

    def define_options(self):
        """Define the sub-command arguments."""
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES
        from colorama import Fore

        # Clinica compulsory arguments
        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')
        clinica_comp.add_argument("group_label",
                                  help='User-defined identifier for the provided group of subjects.')
        clinica_comp.add_argument("orig_input_data",
                                  help='Type of volume-based feature: type '
                                  '\'t1-volume\' to use gray matter maps, '
                                  '\'pet-volume\' to use FDG-PET data or '
                                  '\'custom-pipeline\' to use you own data in CAPS directory '
                                  '(see Wiki for details).',
                                  choices=['t1-volume', 'pet-volume', 'custom-pipeline'])
        clinica_comp.add_argument("subject_visits_with_covariates_tsv",
                                  help='TSV file containing a list of subjects with their sessions and all '
                                       'the covariates and factors needed for the GLM.')
        clinica_comp.add_argument("contrast",
                                  help='Defines the contrast. Must be one of the column names form the TSV file.')

        # Optional arguments (e.g. FWHM)
        optional = self._args.add_argument_group(PIPELINE_CATEGORIES['OPTIONAL'])
        optional.add_argument("-gic", "--group_id_caps", type=str, default=None,
                              help='Name of the group that Clinica needs to use to grab input file.')
        optional.add_argument("-fwhm", "--full_width_at_half_maximum", type=int, default=8,
                              help='Full Width at Half Maximum (FWHM) of the smoothing used in your input file '
                                   '(default: --full_width_at_half_maximum %(default)s).')

        # Optional arguments for custom pipeline
        opt_custom_input = self._args.add_argument_group(
            '%sPipeline options if you selected custom-pipeline%s' % (Fore.BLUE, Fore.RESET)
        )
        opt_custom_input.add_argument("-cf", "--custom_file",
                                      type=str, default=None,
                                      help='Custom file string. Specify filename using * when the subject or session name '
                                      'appears e.g. \'*_task-rest_acq-fdg_pet_space-Ixi549Space_pet.nii.gz\' will grab '
                                      'the corresponding file in all the subjects/sessions. '
                                      'This flag must be specified with the --measure_label flag). '
                                      'See Wiki for an example.')
        opt_custom_input.add_argument("-ml", "--measure_label",
                                      type=str, default=None,
                                      help='Name of the feature type, it will be saved on the CAPS '
                                           '_measure-FEATURE_LABEL key-value association. '
                                           'This flag must be specified with the --custom_file flag). '
                                           'See Wiki for an example.')

        # Clinica standard arguments (e.g. --n_procs)
        self.add_clinica_standard_arguments(add_tsv_flag=False)

        # Advanced arguments (i.e. tricky parameters)
        advanced = self._args.add_argument_group(PIPELINE_CATEGORIES['ADVANCED'])
        advanced.add_argument("-ct", "--cluster_threshold",
                              type=float, default=0.001,
                              help='Threshold to define a cluster in the process of cluster-wise correction '
                                   '(default: --cluster_threshold %(default)s).')

    def run_command(self, args):
        from networkx import Graph
        from .statistics_volume_pipeline import StatisticsVolume
        from clinica.utils.ux import print_end_pipeline, print_crash_files_and_exit
        from clinica.utils.exceptions import ClinicaException

        # Custom pipeline
        if args.orig_input_data == 'custom-pipeline':
            if (args.custom_file is None) or (args.measure_label is None):
                raise ClinicaException('You must set --measure_label and --custom_file flags.')

        parameters = {
            'contrast': args.contrast,
            'orig_input_data': args.orig_input_data,
            'group_label': args.group_id,
            'custom_files': args.custom_files,
            'cluster_threshold': args.cluster_threshold,
            'group_id_caps': args.group_id_caps,
            'full_width_at_half_maximum': args.full_width_at_half_maximum
        }

        pipeline = StatisticsVolume(
            caps_directory=self.absolute_path(args.caps_directory),
            tsv_file=self.absolute_path(args.subject_visits_with_covariates_tsv),
            base_dir=self.absolute_path(args.working_directory),
            parameters=parameters,
            name=self.name
        )

        if args.n_procs:
            exec_pipeline = pipeline.run(plugin='MultiProc',
                                         plugin_args={'n_procs': args.n_procs})
        else:
            exec_pipeline = pipeline.run()

        if isinstance(exec_pipeline, Graph):
            print_end_pipeline(self.name, pipeline.base_dir, pipeline.base_dir_was_specified)
        else:
            print_crash_files_and_exit(args.logname, pipeline.base_dir)
